import sys
sys.path.append("..")
import unittest
import numpy as np
from src.susie.timing_data import TimingData
from astroplan import Observer
from astropy.coordinates import EarthLocation
import astropy.units as u
from src.susie.ephemeris import Ephemeris, LinearModelEphemeris, QuadraticModelEphemeris, PrecessionModelEphemeris, ModelEphemerisFactory
np.random.seed(42)

# Test Data
epochs = np.array([-1640, -1018, -1011, 0, 13, 21, 2017, 2018])
mid_times = np.array([2454515.52496, 2455194.9344, 2455202.57625, 2456305.45536, 2456319.64424, 2456328.37556, 2458506.84758, 2458508.48459])
mid_time_errs = np.array([0.00043, 0.001, 0.0022, 0.00024, 0.00038, 0.00027, 0.00044, 0.00091])
tra_or_occs = np.array(['tra', 'occ', 'occ', 'tra', 'tra', 'tra', 'tra', 'occ'])
tra_or_occ_enum = np.array([0, 1, 1, 0, 0, 0, 0, 1])

# Declaring main base test case object so assertDictAlmostEqual is univerally accessible
class BaseTestCase(unittest.TestCase):
    def assertDictAlmostEqual(self, d1, d2, msg=None, places=6):
        # Helper function used to check if the dictionaries are equal to eachother
        # check if both inputs are dicts
        self.assertIsInstance(d1, dict, 'First argument is not a dictionary')
        self.assertIsInstance(d2, dict, 'Second argument is not a dictionary')
        # check if both inputs have the same keys
        self.assertEqual(d1.keys(), d2.keys())
        # check each key
        for key, value in d1.items():
            if isinstance(value, dict):
                self.assertDictAlmostEqual(d1[key], d2[key], msg=msg)
            elif isinstance(value, np.ndarray):
                self.assertTrue(np.allclose(d1[key], d2[key], rtol=1e-05, atol=1e-08))
            else:
                self.assertAlmostEqual(d1[key], d2[key], places=places, msg=msg)


# Testing the linear model object
class TestLinearModelEphemeris(BaseTestCase):
    def setUp(self):
        # Creates the linear model ephemeris object
        self.ephemeris = LinearModelEphemeris()
        self.epochs = epochs
        self.mid_times = mid_times
        self.mid_time_errs = mid_time_errs
        self.tra_or_occ = tra_or_occs
        self.tra_or_occ_enum = tra_or_occ_enum
        self.T0 = 0.0
        self.P = 1.0

    def test_lin_model_instantiation(self):
        # Tests that ephemeris is an instance of LinearModelEphemeris
        self.assertIsInstance(self.ephemeris, LinearModelEphemeris)
   
    def test_lin_fit(self):
        """Tests that the lin_fit function works.
           Creates a numpy.ndarray[float] with the length of the test data.
        """
        # The expected mid-times returned by the lin-fit equations
        expected_lin_fit = np.array([-1640, -1017.5, -1010.5, 0, 13, 21, 2017, 2018.5])
        actual_lin_fit = self.ephemeris.lin_fit(self.epochs, self.T0, self.P, self.tra_or_occ_enum)
        self.assertTrue(np.allclose(expected_lin_fit, actual_lin_fit, rtol=1e-05, atol=1e-08))

    def test_lin_fit_model(self):
        """Testing that the dictionary parameters of fit model are equal to what they are suppose to be

            Tests the creation of a dictionary named return_data containing the linear fit model data in the order of:
            {
             'period': float,
             'period_err': float,
             'conjunction_time': float,
             'conjunction_time_err': float
            }
        """
        lin_model_fit_result = self.ephemeris.fit_model(self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        lin_model_fit_return_data = {
            'period': 1.0914197876324176,
            'period_err': 3.8445359425767954e-07,
            'conjunction_time': 2456305.4551013056,
            'conjunction_time_err': 0.00034780710703679864
            }
        self.assertDictAlmostEqual(lin_model_fit_result, lin_model_fit_return_data)


# Testing the quadratic model object
class TestQuadraticModelEphemeris(BaseTestCase):
    def setUp(self):
        self.ephemeris = QuadraticModelEphemeris()
        self.epochs = epochs
        self.mid_times = mid_times
        self.mid_time_errs = mid_time_errs
        self.tra_or_occ = tra_or_occs
        self.tra_or_occ_enum = tra_or_occ_enum
        self.T0 = 0.0
        self.P = 1.0
        self.dPdE = 0.0

    def test_quad_fit_instantiation(self):
        # Tests that ephemeris is an instance of QuadraticModelEphemeris
        self.assertIsInstance(self.ephemeris, QuadraticModelEphemeris)

    def test_quad_fit(self):
        """Tests that the quad_fit function works.
           Creates a numpy.ndarray[float] with the length of the test data
        """
        expected_quad_fit = np.array([-1640, -1017.5, -1010.5, 0, 13, 21, 2017, 2018.5])
        actual_quad_fit = self.ephemeris.quad_fit(self.epochs, self.T0, self.P, self.dPdE, self.tra_or_occ_enum)
        self.assertTrue(np.allclose(expected_quad_fit, actual_quad_fit, rtol=1e-05, atol=1e-08))

    def test_quad_fit_model(self):
        """Testing that the dictionary parameters of fit model are equal to what they are suppose to be
            Tests the creation of a dictionary named return_data containing the quadratic fit model data in the order of:
            {  
            'conjunction_time': float,
            'conjunction_time_err': float,
            'period': float,
            'period_err': float,
            'period_change_by_epoch': float,
            'period_change_by_epoch_err': float,
            }
        """
        quad_model_fit_result = self.ephemeris.fit_model(self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        quad_model_return_data = {
            'period': 1.0914201416855518,
            'period_err': 1.0904667249408481e-07,
            'conjunction_time': 2456305.45556849,
            'conjunction_time_err': 0.00010574632139970142,
            'period_change_by_epoch': -1.194938064834552e-09,
            'period_change_by_epoch_err': 1.3310830044505155e-10
            }
        self.assertDictAlmostEqual(quad_model_fit_result, quad_model_return_data)


# Testing the Precession Model Object
class TestPrecessionModelEphemeris(BaseTestCase):
    def setUp(self):
        self.ephemeris = PrecessionModelEphemeris()
        self.epochs = epochs
        self.mid_times = mid_times
        self.mid_time_errs = mid_time_errs
        self.tra_or_occ = tra_or_occs
        self.tra_or_occ_enum = tra_or_occ_enum
        self.T0 = 0.0
        self.P = 1.0
        self.dPdE = 0.0
        self.e = 0.001
        self.w0 = 1.0
        self.dwdE = 0.001

    def test_precession_fit_instantiation(self):
        # Tests that ephemeris is an instance of PrecessionModelEphemeris
        self.assertIsInstance(self.ephemeris, PrecessionModelEphemeris)

    def test_anomalistic_period(self):
        expected_anomalistic_period = 1.00015918
        actual_anomalistic_period = self.ephemeris._anomalistic_period(self.P, self.dwdE)
        self.assertTrue(expected_anomalistic_period, actual_anomalistic_period)
    
    def test_pericenter(self):
        expected_result = np.array([-0.64, -0.018, -0.011, 1, 1.013, 1.021, 3.017, 3.018])
        result = self.ephemeris._pericenter(self.epochs, self.w0, self.dwdE)
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))
        
    def test_precession_fit(self):
        """Tests that the precession_fit function works.
            Creates a numpy.ndarray[int] with the length of the test data
        """
        expected_result = np.array([-1640.000255, -1017.499602, -1010.499602, -1.72010942e-4, 12.99983149, 20.99983365, 2017.000316, 2018.499764])
        result = self.ephemeris.precession_fit(self.epochs, self.T0, self.P, self.e, self.w0, self.dwdE, self.tra_or_occ_enum)
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_precession_fit_model(self):
        """Testing that the dictionary parameters of fit model are equal to what they are suppose to be
            Tests the creation of a dictionary named return_data containing the precession fit model data in the order of:
            {  
            'conjunction_time': float,
            'conjunction_time_err': float,
            'period': float,
            'period_err': float,
            'pericenter_change_by_epoch': float,
            'pericenter_change_by_epoch_err': float,
            'eccentricity': float,
            'eccentricity_err': float,
            'pericenter': float,
            'pericenter_err': float
            }
        """
        prec_model_result = self.ephemeris.fit_model(self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        prec_model_return_data = {
            'period': 1.091433725403284,
            'period_err': 8.897834090343257e-05,
            'conjunction_time': 2456305.48022954,
            'conjunction_time_err': 0.17582298200267588,
            'eccentricity': 0.9722406704173365,
            'eccentricity_err': 48715.865904328704,
            'pericenter': 55488688.58437893,
            'pericenter_err': 1431939.3571539088,
            'pericenter_change_by_epoch': -24013.745815457845,
            'pericenter_change_by_epoch_err': 776.7281997033724
            }
        self.assertDictAlmostEqual(prec_model_result, prec_model_return_data)


class TestModelEphemerisFactory(BaseTestCase):
    def setUp(self):
        # Initialize the Model Ephemeris Factory
        self.ephemeris = ModelEphemerisFactory()
        self.assertIsInstance(self.ephemeris, ModelEphemerisFactory)
        self.epochs = epochs
        self.mid_times = mid_times
        self.mid_time_errs = mid_time_errs
        self.tra_or_occ = tra_or_occs

    def test_model_no_errors_lin(self):
        # Check that a linear model dictionary is returned
        lin_model = self.ephemeris.create_model("linear", self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        return_data = {
            'period': 1.0914197876324176,
            'period_err': 3.8445359425767954e-07,
            'conjunction_time': 2456305.4551013056,
            'conjunction_time_err': 0.00034780710703679864
            }
        self.assertDictAlmostEqual(lin_model, return_data)

    def test_model_no_errors_quad(self):
        # Check that a quadratic model dictionary is returned
        quad_model = self.ephemeris.create_model('quadratic', self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        return_data = {
            'period': 1.0914201416855518,
            'period_err': 1.0904667249408481e-07,
            'conjunction_time': 2456305.45556849,
            'conjunction_time_err': 0.00010574632139970142,
            'period_change_by_epoch': -1.194938064834552e-09,
            'period_change_by_epoch_err': 1.3310830044505155e-10
            }
        self.assertDictAlmostEqual(quad_model, return_data)
    
    def test_model_no_errors_prec(self):
        # Check that a precession model dictionary is returned
        prec_model = self.ephemeris.create_model("precession", self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)
        return_data = {
            'period': 1.091433725403284,
            'period_err': 8.897834090343257e-05,
            'conjunction_time': 2456305.48022954,
            'conjunction_time_err': 0.17582298200267588,
            'eccentricity': 0.9722406704173365,
            'eccentricity_err': 48715.865904328704,
            'pericenter': 55488688.58437893,
            'pericenter_err': 1431939.3571539088,
            'pericenter_change_by_epoch': -24013.745815457845,
            'pericenter_change_by_epoch_err': 776.7281997033724
            }
        self.assertDictAlmostEqual(prec_model, return_data)

    def test_model_errors(self):
        # Check that ValueError is raised if not 'linear', 'quadratic', or 'precession'
        test_model_type = "invalid_model"
        with self.assertRaises(ValueError, msg=f"Invalid model type: {test_model_type}"):
            self.ephemeris.create_model(test_model_type, self.epochs, self.mid_times, self.mid_time_errs, self.tra_or_occ)


class TestEphemeris(BaseTestCase):            
    def setUp(self):
       """Sets up the intantiation of TimingData object and Ephemeris object.

           Runs before every test in the TestEphemeris class
       """
       self.timing_data = TimingData("jd", epochs, mid_times, mid_time_errs, tra_or_occs, time_scale="tdb")
       self.assertIsInstance(self.timing_data, TimingData)
       self.ephemeris = Ephemeris(self.timing_data)

    def test_us_transit_times_instantiation(self):
        """Unsuccessful instantiation of the timing data object within the Ephemeris class

            Need a TimingData object to run Ephemeris
        """
        with self.assertRaises(ValueError, msg="Variable 'timing_data' expected type of object 'TimingData'."):
            self.ephemeris = Ephemeris(None)

    def test_get_model_parameters_linear(self):
        """Tests the creation of the linear model parameters
            With the input of a linear model type, the linear model parameters dictionary is created
            The dictionary is the same one from fit_model in the LinearModelEphemeris
        """
        model_parameters = self.ephemeris._get_model_parameters("linear")
        expected_result = {
            'period': 1.0914197876324176,
            'period_err': 3.8445359425767954e-07,
            'conjunction_time': 2456305.4551013056,
            'conjunction_time_err': 0.00034780710703679864
            }
        self.assertDictAlmostEqual(model_parameters, expected_result)   

    def test_get_model_parameters_quad(self):
        """ Tests the creation of the quadratic model parameters

            With the input of a quadratic model type, the quadratic model parameters dictionary is created
            The dictionary is the same one from fit_model in the QuadraticModelEphemeris
        """
        model_parameters = self.ephemeris._get_model_parameters("quadratic")   
        expected_result = {
            'period': 1.0914201416855518,
            'period_err': 1.0904667249408481e-07,
            'conjunction_time': 2456305.45556849,
            'conjunction_time_err': 0.00010574632139970142,
            'period_change_by_epoch': -1.194938064834552e-09,
            'period_change_by_epoch_err': 1.3310830044505155e-10
            }
        self.assertDictAlmostEqual(model_parameters, expected_result)

    def test_get_model_parameters_prec(self):
        """ Tests the creation of the quadratic model parameters

            With the input of a quadratic model type, the quadratic model parameters dictionary is created
            The dictionary is the same one from fit_model in the QuadraticModelEphemeris
        """
        model_parameters = self.ephemeris._get_model_parameters("precession")   
        expected_result = {
            'period': 1.091433725403284,
            'period_err': 8.897834090343257e-05,
            'conjunction_time': 2456305.48022954,
            'conjunction_time_err': 0.17582298200267588,
            'eccentricity': 0.9722406704173365,
            'eccentricity_err': 48715.865904328704,
            'pericenter': 55488688.58437893,
            'pericenter_err': 1431939.3571539088,
            'pericenter_change_by_epoch': -24013.745815457845,
            'pericenter_change_by_epoch_err': 776.7281997033724
            }
        self.assertDictAlmostEqual(model_parameters, expected_result)

    def test_k_value_linear(self):
        """Tests the correct k value is returned given the linear model type
            The k value for linear model is 2
        """
        expected_result = 2
        result = self.ephemeris._get_k_value("linear")
        self.assertEqual(result, expected_result)
    
    def test_k_value_quad(self):
        """Tests the correct k value is returned given the quadratic model type
            The k value for a quadratic model is 3
        """
        expected_result = 3
        result = self.ephemeris._get_k_value("quadratic")
        self.assertEqual(result, expected_result)

    def test_k_value_prec(self):
        """Tests the correct k value is returned given the quadratic model type
            The k value for a precession model is 5
        """
        expected_result = 5
        result = self.ephemeris._get_k_value("precession")
        self.assertEqual(result, expected_result)

    def test_k_value_err(self):
        """Tests the correct k value is returned given the quadratic model type
            If not 'linear', 'quadratic', or 'precession', will return a ValueError
        """
        test_model_type = "invalid_type"
        with self.assertRaises(ValueError, msg="Only linear, quadratic, and precession models are supported at this time."):
            self.ephemeris._get_k_value(test_model_type)
    
    def test_calc_linear_model_uncertainties(self):
        """ Tests that the correct array of linear uncertainties are produced
            Produces a numpy array with the length of the epochs
            conjunction_time_err = 0.00034780710703679864
            period_err = 3.8445359425767954e-07
        """
        lin_model = self.ephemeris.fit_ephemeris("linear")
        expected_result = np.array([7.20072958e-4, 5.23443252e-4, 5.21435139e-4, 3.47807147e-4, 3.47843055e-4, 3.47900839e-4, 8.49871544e-4, 8.50397753e-4])
        result = self.ephemeris._calc_linear_model_uncertainties(lin_model["conjunction_time_err"], lin_model["period_err"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_calc_quad_model_uncertainties(self):
        """ Tests that the correct array of quadratic uncertainties are produced
            Produces a numpy array with the length of the epochs
            conjunction_time_err = 0.00010574632139970142
            period_err = 1.0904667249408481e-07
            period_change_by_epoch_err = 1.3310830044505155e-10
        """
        quad_model = self.ephemeris.fit_ephemeris("quadratic")
        expected_result = np.array([2.74239386e-4, 1.68078507e-4, 1.67188748e-4, 1.05746321e-4, 1.05755824e-4, 1.05771118e-4, 3.64514509e-4, 3.64812697e-4])
        result = self.ephemeris._calc_quadratic_model_uncertainties(quad_model["conjunction_time_err"], quad_model["period_err"], quad_model["period_change_by_epoch_err"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_calc_linear_ephemeris(self):
        """ Tests that the correct linear model data is produced
            The model data is a numpy array of calcuated mid transit times
            conjunction time = 2456305.4551013056
            period = 1.0914197876324176
        """
        lin_model = self.ephemeris.fit_ephemeris("linear")
        expected_result = np.array([2454515.527, 2455194.935, 2455202.575, 2456305.455, 2456319.644, 2456328.375, 2458506.849, 2458508.486])
        result = self.ephemeris._calc_linear_ephemeris(self.timing_data.epochs, lin_model["conjunction_time"], lin_model["period"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_calc_quadratic_ephemeris(self):
        """ Tests that the correct quadratic model data is produced
            The model data is a numpy array of calcuated mid transit times
            conjunction_time = 2456305.45556849
            period = 1.0914201416855518
            period change by epoch = -1.194938064834552e-09
        """
        quad_model = self.ephemeris.fit_ephemeris("quadratic")
        expected_result = np.array([2454515.525, 2455194.935, 2455202.575, 2456305.456, 2456319.644, 2456328.375, 2458506.848, 2458508.485])
        result = self.ephemeris._calc_quadratic_ephemeris(self.timing_data.epochs, quad_model["conjunction_time"], quad_model["period"], quad_model["period_change_by_epoch"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_calc_precession_ephemeris(self):
        """ Tests that the correct precession model data is produced
            The model data is a numpy array of calcuated mid transit times
            conjunction_time = 2456305.48022954
            period = 1.091433725403284
            eccentricity = 0.9722406704173365
            pericenter = 55488688.58437893
            pericenter change by epoch = -24013.745815457845
        """
        prec_model = self.ephemeris.fit_ephemeris("precession")
        expected_result = np.array([2454515.529, 2455194.401, 2455202.041, 2456305.48, 2456319.669, 2456328.4, 2458506.902, 2458507.994])
        result = self.ephemeris._calc_precession_ephemeris(self.timing_data.epochs, prec_model["conjunction_time"], prec_model["period"], prec_model["eccentricity"], prec_model["pericenter"], prec_model["pericenter_change_by_epoch"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_calc_chi_squared_linear(self):
        """ Tests the calculated chi squared value
            The linear chi squared value is a float that is calculated with the model data produced by test_calc_linear_ephemeris
            linear data = [2454515.52664959, 2455194.93546739, 2455202.5754059, 2456305.45510131, 
                           2456319.64355855, 2456328.37491685, 2458506.84881296, 2458508.48594264]
        """
        lin_model = self.ephemeris.fit_ephemeris("linear")
        expected_result = 36.83922943
        result = self.ephemeris._calc_chi_squared(lin_model["model_data"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))   

    def test_calc_chi_squared_quad(self):
        """ Tests the calculated chi squared value
            The quadratic chi squared value is a float that is calculated with the model data produced by test_calc_quadratic_ephemeris 
            quad data = [2454515.52492917, 2455194.93495515, 2455202.57490463, 2456305.45556849,
                         2456319.64403023, 2456328.3753912, 2458506.84756359, 2458508.48469139]
        """
        quad_model = self.ephemeris.fit_ephemeris("quadratic")
        expected_result = 2.15135181
        result = self.ephemeris._calc_chi_squared(quad_model["model_data"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))   
    
    def test_calc_chi_squared_prec(self):
        """ Tests the calculated chi squared value
            The precession chi squared value is a float that is calculated with the model data produced by test_calc_precession_ephemeris 
            prec data = [2454515.52900788, 2455194.40083254, 2455202.04080698, 2456305.48015054,
                         2456319.66889067, 2456328.40025227, 2458506.90196704, 2458507.99371183]
        """
        prec_model = self.ephemeris.fit_ephemeris("precession")
        expected_result = 673519.3095
        result = self.ephemeris._calc_chi_squared(prec_model["model_data"])
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

    def test_subract_linear_parameters(self):
        """ Tests subtracting linear terms from the model data
        data = [2454515.52492917, 2455194.93495515, 2455202.57490463, 2456305.45556849, 
                2456319.64403023, 2456328.3753912, 2458506.84756359, 2458508.48469139]
        conjunction time = 2456305.45556849
        period = 1.0914201416855518
        tra or occ = ['tra', 'occ', 'occ', 'tra', 'tra', 'tra', 'tra', 'occ']
        """
        quad_data = self.ephemeris.fit_ephemeris("quadratic")
        expected_result = np.array([-0.001606955695, -6.19174951e-4, -6.1068675e-4, 0, -1.01912175e-7, -2.6539659e-7, -0.002430679758, -0.002433092286])
        result = self.ephemeris._subtract_linear_parameters(quad_data["model_data"], quad_data["conjunction_time"], quad_data["period"], self.ephemeris.timing_data.epochs, self.ephemeris.timing_data.tra_or_occ)
        self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))
    
    def test_fit_ephemeris_linear(self):
        """ Tests that the linear model type produces the linear model parameters with the linear model type and linear model data included

            Uses the test_get_model_parameters_linear and test_calc_linear_ephemeris to produce a dictionary with:
            {
            'period': float,  
            'period_err': float,
            'conjunction_time': float,
            'conjunction_time_err':  float,
            'model_type': 'linear', 
            'model_data': np.array
        }
        """
        model_parameters_linear = {
            'period': 1.0914197876324176,
            'period_err': 3.8445359425767954e-07,
            'conjunction_time': 2456305.4551013056,
            'conjunction_time_err': 0.00034780710703679864,
            'model_type': 'linear',
            'model_data': [2454515.52664959, 2455194.93546739, 2455202.5754059 ,
                    2456305.45510131, 2456319.64355855, 2456328.37491685,
                    2458506.84881296, 2458508.48594264]}
        result = self.ephemeris.fit_ephemeris("linear")
        self.assertDictAlmostEqual(result, model_parameters_linear)

    def test_fit_ephemeris_quad(self):
        """ Tests that the quadratic model type produces the quadratic model parameters with the quadratic model type and quadratic model data included

            Uses the test_get_model_parameters_linear and test_calc_linear_ephemeris to produce a dictionary with:
            {
            'period': float,  
            'period_err': float,
            'conjunction_time': float,
            'conjunction_time_err':  float,
            'period_change_by_epoch': float,
            'period_change_by_epoch_err': float,
            'model_type': 'quadratic', 
            'model_data': np.array
        }
        """
        model_parameters_quad = {
            'period': 1.0914201416855518,
            'period_err': 1.0904667249408481e-07,
            'conjunction_time': 2456305.45556849,
            'conjunction_time_err': 0.00010574632139970142,
            'period_change_by_epoch': -1.194938064834552e-09,
            'period_change_by_epoch_err': 1.3310830044505155e-10,
            'model_type': 'quadratic',
            'model_data': [2454515.52492917, 2455194.93495515, 2455202.57490463,
                    2456305.45556849, 2456319.64403023, 2456328.3753912 ,
                    2458506.84756359, 2458508.48469139]}
        result = self.ephemeris.fit_ephemeris("quadratic")
        self.assertDictAlmostEqual(result, model_parameters_quad)

    def test_fit_ephemeris_prec(self):
        """ Tests that the precession model type produces the precession model parameters with the precession model type and precession model data included

            Produces a dictionary with:
            {
            'period': float,  
            'period_err': float,
            'conjunction_time': float,
            'conjunction_time_err':  float,
            'eccentricity': float, 
            'eccentricity_err': float,
            'pericenter': float,
            'pericenter_err': float,
            'pericenter_change_by_epoch': float,
            'pericenter_change_by_epoch_err': float,
            'model_type': 'precession', 
            'model_data': np.array
        }
        """
        model_parameters_prec = {
            'period': 1.091433725403284,
            'period_err': 8.897834090343257e-05,
            'conjunction_time': 2456305.48022954,
            'conjunction_time_err': 0.17582298200267588,
            'eccentricity': 0.9722406704173365,
            'eccentricity_err': 48715.865904328704,
            'pericenter': 55488688.58437893,
            'pericenter_err': 1431939.3571539088,
            'pericenter_change_by_epoch': -24013.745815457845,
            'pericenter_change_by_epoch_err': 776.7281997033724,
            'model_type': 'precession',
            'model_data': [2454515.52900788, 2455194.40083254, 2455202.04080698,
                    2456305.48015054, 2456319.66889067, 2456328.40025227,
                    2458506.90196704, 2458507.99371183]}
        result = self.ephemeris.fit_ephemeris("precession")
        self.assertDictAlmostEqual(result, model_parameters_prec)
    
    def test_get_ephemeris_uncertainites_model_type_err(self):
        """ Unsuccessful test to calculate uncertainties
            Model type is needed
        """
        # Just using random data
        model_parameters = {
            'period': 0.0,
            'period_err': 0.0,
            'conjunction_time': 0.0,
            'conjunction_time_err': 0.0,
            'model_data': np.array([0.0, 0.0])
        }
        with self.assertRaises(KeyError, msg="Cannot find model type in model data. Please run the fit_ephemeris method to return ephemeris fit parameters."):
            self.ephemeris.get_ephemeris_uncertainties(model_parameters)

    def test_get_ephemeris_uncertainties_lin_err(self):
        """ Unsuccessful test to calculate uncertainties
            Period error and conjunction time error values are needed
        """
        model_parameters = {
            'period': 1.0,
            'conjunction_time': 0.0,
            'model_type': 'linear',
            'model_data': np.array([0.0, 0.0])
        }
        with self.assertRaises(KeyError, msg="Cannot find conjunction time and period errors in model data. Please run the fit_ephemeris method with 'linear' model_type to return ephemeris fit parameters."):
            self.ephemeris.get_ephemeris_uncertainties(model_parameters)

    def test_get_ephemeris_uncertainties_quad_err(self):
        """ Unsuccessful test to calculate uncertainties
            Conjunction time error, period error and period change by epoch error is needed
        """
        model_parameters = {
        'period': 1.0,
        'conjunction_time': 0.0,
        'period_change_by_epoch': 0.0,
        'model_type': 'quadratic',
        'model_data': np.array([0.0, 0.0])
        }
        with self.assertRaises(KeyError, msg="Cannot find conjunction time, period, and/or period change by epoch errors in model data. Please run the fit_ephemeris method with 'quadratic' model_type to return ephemeris fit parameters."):
            self.ephemeris.get_ephemeris_uncertainties(model_parameters)

    def test_get_ephemeris_uncertainties_prec_err(self):
        """ Unsuccessful test to calculate uncertainties
            Conjunction time error, period error and period change by epoch error is needed
        """
        model_parameters = {
        'period': 1.0,
        'conjunction_time': 0.0,
        'period_change_by_epoch': 0.0,
        'model_type': 'precession',
        'model_data': np.array([0.0, 0.0])
        }
        with self.assertRaises(KeyError, msg="Cannot find conjunction time, period, eccentricity, pericenter, and/or pericenter change by epoch and/or their respective errors in model data. Please run the fit_ephemeris method with 'precession' model_type to return ephemeris fit parameters."):
            self.ephemeris.get_ephemeris_uncertainties(model_parameters)

    # def test_get_ephemeris_uncertainites_linear(self):
    #     """ Sucessful test to calculate linear uncertainties
    #         Expected result is the numpy array produced by test_calc_linear_model_uncertainties
    #     """
    #     model_parameters_linear = {
    #         'period': 1.0904734088754364,
    #         'period_err': 0.0006807481006299065,
    #         'conjunction_time': 2454515.423966982,
    #         'conjunction_time_err': 0.23692092991744518,
    #         'model_type': 'linear',
    #         'model_data': np.array([2454515.42396698, 2454836.5683859 , 2454840.38504283,
    #                 2455140.81046697])
    #     }
    #     expected_result = np.array([0.23692092, 0.31036088, 0.31190525, 0.45667354])
    #     self.ephemeris.get_ephemeris_uncertainties(model_parameters_linear)
    #     results = self.ephemeris.get_ephemeris_uncertainties(model_parameters_linear)
    #     self.assertTrue(np.allclose(expected_result, results, rtol=1e-05, atol=1e-08)) 
         
    # def test_get_ephemeris_uncertainites_quad(self):
    #     """ Sucessful test to calculate quadratic uncertainties

    #         Expected result is the numpy array produced by test_calc_quadratic_model_uncertaintie
    #     """
    #     model_parameters_quad = {
    #     'period': 1.0892663209112947,
    #     'period_err': 0.002368690041166098,
    #     'conjunction_time': 2454515.5241231285,
    #     'conjunction_time_err': 0.3467430587812461,
    #     'period_change_by_epoch': 4.223712653342504e-06,
    #     'period_change_by_epoch_err': 7.742732700893123e-06,
    #     'model_type': 'quadratic',
    #     'model_data': np.array([2454515.52412313, 2454836.49559505, 2454840.31302805,2455140.91174185])
    #     }
    #     expected_result = np.array([0.34674306,0.84783352,0.85829843,1.89241887])
    #     self.ephemeris.get_ephemeris_uncertainties(model_parameters_quad)
    #     results = self.ephemeris.get_ephemeris_uncertainties(model_parameters_quad)
    #     self.assertTrue(np.allclose(expected_result, results, rtol=1e-05, atol=1e-08)) 
         
    # def test_calc_bic_lin(self):
    #     """ Tests the calculation of the linear bic

    #         Uses the linear k value and linear chi squared value
    #     """
    #     model_parameters_linear = {
    #         'period': 1.0904734088754364,
    #         'period_err': 0.0006807481006299065,
    #         'conjunction_time': 2454515.423966982,
    #         'conjunction_time_err': 0.23692092991744518,
    #         'model_type': 'linear',
    #         'model_data': np.array([2454515.42396698, 2454836.5683859 , 2454840.38504283,
    #                 2455140.81046697])
    #     }
    #     # k_value = 2
    #     expected_result = 843769.0757319723
    #     result = self.ephemeris.calc_bic(model_parameters_linear)
    #     self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))   
        
    # def test_calc_bic_quad(self):
    #     """ Tests the calculation of the quadratic bic

    #         Uses the quadratic k value and quadratic chi squared value
    #     """
    #     model_parameters_quad = {
    #     'period': 1.0892663209112947,
    #     'period_err': 0.002368690041166098,
    #     'conjunction_time': 2454515.5241231285,
    #     'conjunction_time_err': 0.3467430587812461,
    #     'period_change_by_epoch': 4.223712653342504e-06,
    #     'period_change_by_epoch_err': 7.742732700893123e-06,
    #     'model_type': 'quadratic',
    #     'model_data': np.array([2454515.52412313, 2454836.49559505, 2454840.31302805,2455140.91174185])
    #     }
    #     # k_value = 3
    #     expected_result = 650255.7398105409
    #     result = self.ephemeris.calc_bic(model_parameters_quad)
    #     self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08)) 
    

    # def test_calc_delta_bic(self):
    #     """ Tests the calulation of the delta bic

    #         Uses both the quadratic bic and linear bic
    #     """
    #     expected_result = 193513.33592143143
    #     result = self.ephemeris.calc_delta_bic(model_1, model_2)
    #     print(f"\n\n\n MEEP \n {expected_result}, {result} \n\n\n")
    #     self.assertTrue(expected_result, result)


# class TestPrecessionEphemeris(unittest.TestCase):        
#     def setUp(self):
#        """Sets up the intantiation of TimingData object and Ephemeris object.

#            Runs before every test in the TestEphemeris class
#        """
#        self.timing_data = TimingData('jd', test_epochs_precession, test_mtts_precession, test_mtts_err_precession, test_tra_or_occ_precession, time_scale='tdb')
#        self.assertIsInstance(self.timing_data, TimingData)
#        self.ephemeris = Ephemeris(self.timing_data)


#     def test_suc_t0_calc(self):
#         expected_result = 0.0000102238439463
#         result = self.ephemeris._calc_t0_model_uncertainty(test_T0_err_pre)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))   
    
#     def test_suc_e_calc(self):
#         expected_result = 0.167733834964
#         result = self.ephemeris._calc_eccentricity_model_uncertainty(test_P_pre, test_dwdE, test_w, -1640, test_e_err)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))   
    
#     def test_suc_w_calc(self):
#         expected_result = 0.543545754247
#         result = self.ephemeris._calc_pericenter_model_uncertainty(test_e, test_P_pre, test_dwdE, test_w, -1640, test_w_err)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))     
    
#     def test_suc_dwdE_trans_calc(self):
#         expected_result = 0.00805534737998
#         result = self.ephemeris._calc_change_in_pericenter_transit_model_uncertainty(test_e, test_P_pre, test_dwdE, test_w, -1640, test_dwdE_err)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08)) 

#     def test_suc_dwdE_occ_calc(self):
#         expected_result = 0.00805539085732
#         result = self.ephemeris._calc_change_in_pericenter_transit_model_uncertainty(test_e, test_P_pre, test_dwdE, test_w, -1640, test_dwdE_err)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))

#     def test_suc_P_trans_calc(self):
#         expected_result = 0.000017849816617
#         result = self.ephemeris._calc_period_transit_model_uncertainty(test_e, test_dwdE, test_w, -1640, test_P_err_pre)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))  

#     def test_suc_P_occ_calc(self):
#         expected_result = 0.0000178498155019
#         result = self.ephemeris._calc_period_occ_model_uncertainty(test_e, test_dwdE, test_w, -1640, test_P_err_pre)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))    

#     def test_final_precession_uncertainty(self):
#         test_model_params = {'period': 1.091423398620008,
#             'period_err': 2.5761611591762386e-06,
#             'conjunction_time': 2456305.4591182857,
#             'conjunction_time_err': 0.0031974746201258552,
#             'eccentricity': 0.709249384779316,
#             'eccentricity_err': 21948.154831900636,
#             'pericenter': 201043977.85898417,
#             'pericenter_err': 594028.6944042204,
#             'pericenter_change_by_epoch': -116462.52579119639,
#             'pericenter_change_by_epoch_err': 44.094731428936605
#             }
#         expected_result = np.array([0.848152704559,0.500184477109,6.678399994, 7.03100035237, 7.43172845908, 0.517645163004, 7.30881685475])
#         result = self.ephemeris._calc_precession_model_uncertainties(test_model_params)
#         self.assertTrue(np.allclose(expected_result, result, rtol=1e-05, atol=1e-08))    


#     # Astroplan Test
#     # observer obj
#     def test_obs_obj_lat_long(self):
#         #  tests that if (lat, long) is passed in creates  observer obj
#         boise_state = (-116.2010,43.6023, 821)
#         timezone = 'US/Mountain'
#         observer = self.ephemeris._create_observer_obj(timezone, coords = boise_state)
#         self.assertEqual(observer.location.lat.deg, 43.6023)
#         self.assertEqual(observer.location.lon.deg, -116.2010)
    
#     def test_obs_obj_name(self):
#         # tests that if a name is passed in uses observer.at_site
#         name = 'Subaru'
#         timezone = 'US/Hawaii'
#         empty_coords = (None, None, None)
#         observer = self.ephemeris._create_observer_obj(timezone, coords=empty_coords, name=name)
#         self.assertEqual(observer.name, "Subaru")

#     def test_suc_return_obs_obj(self):
#         # tests for type astroplan.observer , name = str, coord = (float, float)
#         name = 'Subaru'
#         timezone = 'US/Hawaii'
#         empty_coords = (None, None, None)
#         observer = self.ephemeris._create_observer_obj(timezone, coords=empty_coords, name=name)
#         self.assertTrue(all(isinstance(value, str) for value in  observer.name))
        
#     def test_suc_return_obs_obj_float(self):
#         boise_state = (-116.2010,43.6023, 821.0)
#         timezone = 'US/Mountain'
#         observer = self.ephemeris._create_observer_obj(timezone, coords = boise_state)
#         self.assertTrue((isinstance(observer.location.lat.deg, (float, np.float64))))
#         self.assertTrue((isinstance(observer.location.lon.deg, (float, np.float64))))
#         # self.assertTrue((isinstance(observer.location.height (float, np.float64))))

#     def test_obs_obj_value_err(self):
#         # tests for value error if no name or lat or long
#         timezone = 'US/Hawaii'
#         empty_coords = (None, None, None)
#         with self.assertRaises(ValueError, msg="Observatory location must be specified with either (1) a site name specified by astropy.coordinates.EarthLocation.get_site_names() or (2) latitude and longitude in degrees as accepted by astropy.coordinates.Latitude and astropy.coordinates.Latitude."):
#             self.ephemeris._create_observer_obj(timezone, coords = empty_coords, name= None)

#     # target obj
#     def test_target_obj_ra_dec(self):
#         #  tests that if coords = (ra,dec) is passed in creates a fixed target with the ra and dec 
#         # check the ra and dec is returned
#         tres_3 = (268.0291,37.54633)
#         target = self.ephemeris._create_target_obj(coords = tres_3)
#         self.assertEqual(target.ra.deg, 268.0291)
#         self.assertEqual(target.dec.deg, 37.54633)
    
#     def test_target_obj_name(self):
#         # tests that if a name is passed in uses fixedTarget.from_name
#         tres_3 = (None,None)
#         target = self.ephemeris._create_target_obj(coords = tres_3, name='TrES-3b')
#         self.assertEqual(target.name, "TrES-3b")
    
#     def test_suc_return_target_obj(self):
#         # tests for type astroplan.FixedTarget , name = str, coord = (float, float)
#         name = 'TrES-3b'
#         empty_coords = (None, None)
#         target = self.ephemeris._create_target_obj(coords=empty_coords, name=name)
#         self.assertTrue(all(isinstance(value, str) for value in  target.name))
    
#     def test_suc_return_obj_float(self):
#         tres_3 = (268.0291,37.54633)
#         target = self.ephemeris._create_target_obj(coords = tres_3)
#         self.assertTrue((isinstance(target.ra.deg, (float, np.float64))))
#         self.assertTrue((isinstance(target.dec.deg, (float, np.float64))))
      
#     def test_target_obj_value_err(self):
#         # tests for value error if no name or ra or dec
#         empty_coords = (None, None, None)
#         with self.assertRaises(ValueError, msg="Object location must be specified with either (1) an valid object name or (2) right ascension and declination in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec."):
#             self.ephemeris._create_target_obj(coords = empty_coords, name= None)
    
#     # query NASA
#     def test_input_value_err(self):
#         # tests for error if no coords or name
#         with self.assertRaises(ValueError, msg="Object must be specified with either (1) a recognized object name in the NASA Exoplanet Archive or (2) right ascension and declination in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec."):
#             self.ephemeris._query_nasa_exoplanet_archive(None, ra=None, dec=None, select_query=None)
    
#     def test_nothing_found_name_value_err(self):
#         # tests for if nothing is found in the query for the name
#         with self.assertRaises(ValueError, msg=f"Nothing found for {'Earth'} in the NASA Exoplanet Archive. Please check that your object is accepted and contains data on the archive's homepage."):
#             self.ephemeris._query_nasa_exoplanet_archive("Earth", ra=None, dec=None, select_query=None)
   
#     def test_nothing_found_ra_dec_value_err(self):
#         # tests for if nothing is found in the query for the ra and dec
#         with self.assertRaises(ValueError, msg=f"Nothing found for the coordinates {0}, {0} in the NASA Exoplanet Archive. Please check that your values are correct and are in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec."):
#             self.ephemeris._query_nasa_exoplanet_archive('ground', ra=0, dec=0, select_query=None)
   
#     def test_len_over_zero(self):
#         # tests the obj data length is not zero
#         pass

#     # eclipse duration
#     def test_eclipse_duration(self):
#         # tests calc eclipse duration
#         # trES_3 values
#         test_k = 0.1655
#         test_P = 1.30618581
#         test_a = 0.02282
#         test_b = 0.840
#         test_i = 81.85
#         test_R_star_a = 1/5.926
#         test_R_star = 1/(5.926 * (1/0.02282))
#         test_R_planet = 14.975
#         transit_duration = 3.296814064
#         result = self.ephemeris._calc_eclipse_duration(test_P, test_R_star, test_R_planet, test_a, test_b, test_i)
#         self.assertEqual(transit_duration, result)

#     # eclipse system params
#     def test_eclipse_params_type(self):
#         # tests type float 
#         pass
#     def test_eclipse_params_not_empty(self):
#         # tests that theres values but not using any values
#         pass
    
#     # observing schedule
#     def test_primary_eclipse(self):
#         # tests for the primary eclipse time using our data 
#         # is the last eclipse time
#         pass

#     def test_orbital_period(self):
#         # tests its a float, units and value
#         # from model period
#         pass


    if __name__ == '__main__':
            unittest.main()


import numpy as np
from astropy import time
from astropy import coordinates as coord
from astropy import units as u
import logging

logger = logging.getLogger("lumberjack")

class TimingData():
    """Represents timing mid-point data over observations. Holds data to be accessed by :class:`Ephemeris`.

    This object processes, formats, and holds user data to be passed to the :class:`Ephemeris` object.
    Users can input observational data as lists of transit and/or occultation mid-times and corresponding
    epochs and (if available) mid-time uncertainties. 
    
    Timing conversions are applied to ensure that all data is processed correctly and users are aware of 
    timing formats and scales, which can give rise to false calculations in our metrics. If data is not specified
    to be in Barycentric Julian Date format with the TDB time scale, timing data will be corrected for barycentric
    light travel time using the Astropy Time utilities. If correction is needed, users will be required to provide
    additional information on the observed object.
    
    If mid time uncertainties are not provided, we will generate placeholder values of 1.

    Our implementations rely on Numpy functions. This object implements checks to ensure that data are stored in 
    Numpy arrays and are of correct data types. The appropriate Type or Value Error is raised if there are any 
    issues.

    All timing data arrays will be sorted by ascending epoch. Epochs are shifted to start at zero by subtracting 
    the minimum number from each value.

    Parameters
    ------------
        time_format: str 
            An abbreviation of the data's time system. Abbreviations for systems can be found on [Astropy's 
            Time documentation](https://docs.astropy.org/en/stable/time/#id3).
        epochs: numpy.ndarray[int]
            List of orbit number reference points for timing observations
        mid_times: numpy.ndarray[float]
            List of observed timing mid points corresponding with epochs, in timing units given by time_format.
        mid_time_uncertainties: Optional(numpy.ndarray[float])
            List of uncertainties corresponding to timing mid points, in timing units given by time_format. If 
            given None, will be replaced with array of 1's with same shape as `mid_times`.
        tra_or_occ: Optional(numpy.ndarray[str])
            List of either `tra` or `occ` which specifies if observational data was taken from a transit or an 
            occultation.
        time_scale: Optional(str)
            An abbreviation of the data's time scale. Abbreviations for scales can be found on [Astropy's Time 
            documentation](https://docs.astropy.org/en/stable/time/#id6).
        object_ra: Optional(float)
            The right ascension of observed object represented by data, in degrees
        object_dec: Optional(float)
            The declination of observed object represented by data, in degrees
        observatory_lon: Optional(float)
            The longitude of observatory data was collected from, in degrees
        observatory_lat: Optional(float) 
            The latitude of observatory data was collected from, in degrees
    
    Raises
    ------
        TypeError: 
            - If ``epochs``, ``mid_times``, ``mid_time_uncertainties``, and/or ``tra_or_occs`` are not Numpy arrays.
            - If ``epochs`` contain any non-int values.
            - If ``mid_times`` and/or ``mid_time_uncertainties`` contain any non-float values.

        ValueError:
            - If ``epochs``, ``mid_times``, ``mid_time_uncertainties``, and/or ``tra_or_occs`` do not have the same 
            amount of data (the arrays do not have the same shape).
            - If ``tra_or_occ`` contains any values that are not 'tra' or 'occ'.
            - If ``mid_times`` or ``mid_time_uncertainties`` contain any NaN values.
            - If ``mid_time_uncertainties`` contain any negative or zero values.
    """
    def __init__(self, time_format, epochs, mid_times, mid_time_uncertainties=None, tra_or_occ=None, time_scale=None, object_ra=None, object_dec=None, observatory_lon=None, observatory_lat=None):
        # Configure logging to remove "root" prefix
        self._configure_logging()
        self.epochs = epochs
        self.mid_times = mid_times
        self.mid_time_uncertainties = mid_time_uncertainties
        if tra_or_occ is None:
            # Create list of just "tra"
            tra_or_occ = np.array(["tra" for el in self.epochs])
        self.tra_or_occ = tra_or_occ
        # Check that timing system and scale are JD and TDB
        if time_format != "jd" or time_scale != "tdb":
            # If not correct time format and scale, create time objects and run corrections
            logging.warning(f"Recieved time format {time_format} and time scale {time_scale}. Correcting all times to BJD timing system with TDB time scale. If no time scale is given, default is automatically assigned to UTC. If this is incorrect, please set the time format and time scale for TransitTime object.")
            # Set timing data to None for now
            self.mid_times = self._convert_times(mid_times, time_format, time_scale, (object_ra, object_dec), (observatory_lon, observatory_lat), warn=True)
            if mid_time_uncertainties is not None:
                self.mid_time_uncertainties = self._convert_timing_uncertainties(mid_times, mid_time_uncertainties, time_format, time_scale, (object_ra, object_dec), (observatory_lon, observatory_lat))
        if mid_time_uncertainties is None:
            # If no uncertainties provided, make an array of 1s in the same shape of epochs
            logging.warning(f"Recieved value of {mid_time_uncertainties} for mid time uncertainties. Auto-populating placeholder values of 0.001 for uncertainties.")
            self.mid_time_uncertainties = np.full(self.epochs.shape, 0.001)
        # Call validation function
        self._validate()
        # Once every array is populated, make sure you sort by ascending epoch
        self._sort_data_arrays()

    def _calc_barycentric_time(self, time_obj, obj_location):
        """Function to correct non-barycentric time formats to Barycentric Julian Date in TDB time scale.

        This function will run under the given circumstances:

        - If the timing format provided is not JD (``time_format`` does not equal "jd")
        - If the timing scale provided is not TDB (``time_scale`` does not equal "tdb")
        - If the timing scale is not provided

        Calculates the light travel time for the given Astropy timing object and adds the light 
        travel time to each original value in the given timing data. If the given Astropy timing object time data 
        contains a list of 1s, which means this is placeholder timing uncertainty data, no timing correction will 
        be applied as this is not real data. If the timing correction proceeds, the ``light_travel_time`` function 
        from Astropy will be applied and added to the original timing data. Timing data corrected for Barycentric  
        light travel time will be returned.

        Parameters
        ----------
            time_obj : numpy.ndarray[float]
               List of timing data to be corrected to the Barycentric Julian Date time format in the TDB time scale.
            obj_location : Astropy.coordinates.SkyCoord obj
                The RA and Dec in degrees of the object being observed, stored as an Astropy coordinates.SkyCoord object.
            obs_location : Astropy.coordinates.EarthLocation obj
                The longitude and latitude in degrees of the site of observation, stored as an Astropy coordinates.EarthLocation object. 
                If None given, uses gravitational center of Earth at North Pole.
       
        Returns
        -------
            time_obj.value : numpy.ndarray[float]
                Returned only if these are placeholder uncertainties and no correction is needed.
            corrected_time_vals : numpy.ndarray[float]
                List of mid times corrected to Barycentric Julian Date time format with TDB time scale.
        """
        # If given uncertainties, check they are actual values and not placeholders vals of 1
        # If they are placeholder vals, no correction needed, just return array of 1s
        ltt_bary = time_obj.light_travel_time(obj_location)
        corrected_time_vals = (time_obj.tdb+ltt_bary).value
        return corrected_time_vals
    
    def _configure_logging(self):
        logging.basicConfig(format="%(levelname)s: %(message)s")

    def _convert_timing_uncertainties(self, mid_times, mid_time_uncertainties, format, scale, obj_coords, obs_coords):
        """Calculates the converted mid-time uncertainties.

        Calculates the new converted timing uncertainties by calculating the upper and lower limits
        of each mid-time, converting the time formats and scales of the upper and lower limits, then
        subtracting the converted mid time from the limits and taking the square root of the sum of 
        squares for the final resulting mid-time uncertainty.

        This function will ONLY run if the timing format and/or scale needs to be converted and mid-time
        uncertainties are given. If no mid-time uncertainties are given, this calculation will not run and
        placeholder uncertainty values will be generated instead.

        Parameters
        ----------
            mid_times: numpy.ndarray[float]
                List of observed timing mid-points corresponding with epochs.
            mid_time_uncertainties: numpy.ndarray[float]
                List of uncertainties corresponding to timing mid-points.
            format: str
                A valid Astropy abbreviation of the data's time system.
            scale: str
                A valid Astropy abbreviation of the data's time scale.
            obj_coords: (float, float)
                Tuple of the right ascension and declination in degrees of the object being observed.
            obs_coords: (float, float)    
                Tuple of the longitude and latitude in degrees of the site of observation.

        Returns
        -------
            unc: np.ndarray[float]
                An array of timing uncertainty data converted to Barycentric Julian Date timing format and scale (Astropy JD format, TDB scale).
        """
        # create time objects from upper and lower vals
        mid_times_obj = time.Time(mid_times, format=format, scale=scale)
        upper_times_obj = time.Time(mid_times + mid_time_uncertainties, format=format, scale=scale)
        lower_times_obj = time.Time(mid_times - mid_time_uncertainties, format=format, scale=scale)
        # convert the format mid_times, up and down errs
        mid_times_converted = self._convert_times(mid_times_obj, format, scale, obj_coords, obs_coords)
        upper_times_converted = self._convert_times(upper_times_obj, format, scale, obj_coords, obs_coords)
        lower_times_converted = self._convert_times(lower_times_obj, format, scale, obj_coords, obs_coords)
        # subtract up and down errs
        upper_diffs = upper_times_converted - mid_times_converted
        lower_diffs = mid_times_converted - lower_times_converted
        unc = np.sqrt((upper_diffs**2) + (lower_diffs**2))
        return unc

    def _convert_times(self, time_data, format, scale, obj_coords, obs_coords, warn=False):
        """Validates object and observatory information and populates Astropy objects for barycentric light travel time correction.

        Checks that object coordinates (right ascension and declination) and are of correct types. If correct object 
        coordinates are given, will create an Astropy SkyCoord object. Checks that observatory coordinates (longitude 
        and latitude) are given and of correct types. If given, will populate an Astropy EarthLocation object. If not
        given, will populate Astropy EarthLocation with gravitational center of Earth at North Pole. Passes the validated
        SkyCoord, EarthLocation, and Time objects to the `_calc_barycentric_time` correction function to convert times
        to BJD TDB timing format and scale.

        Parameters
        ----------
            time_data: np.ndarray[float]
                An array of timing data values. This will either be mid times or the mid time uncertainties.
            format: str
                A valid Astropy abbreviation of the data's time system.
            scale: str
                A valid Astropy abbreviation of the data's time scale.
            obj_coords: (float, float)
                Tuple of the right ascension and declination in degrees of the object being observed.
            obs_coords: (float, float)    
                Tuple of the longitude and latitude in degrees of the site of observation.
            warn: Boolean
                If True, will raise warnings to the user.

        Returns
        -------
            An array of timing data converted to Barycentric Julian Date timing format and scale (Astropy JD format, TDB scale).

        Raises
        ------
            ValueError:
                Error if None recieved for object_ra or object_dec.
            Warning:
                Warning if no observatory coordinates are given.
                Warning notifying user that barycentric light travel time correction is taking place with the given
                information.
        """
        # Get observatory and object location
        obs_location = self._get_obs_location(obs_coords, warn)
        obj_location = self._get_obj_location(obj_coords)
        if warn:
            logging.warning(f"Using ICRS coordinates in degrees of RA and Dec {round(obj_location.ra.value, 2), round(obj_location.dec.value, 2)} for time correction. Using geodetic Earth coordinates in degrees of longitude and latitude {round(obs_location.lon.value, 2), round(obs_location.lat.value, 2)} for time correction.")
        # Create time object and convert format to JD
        time_obj = time.Time(time_data, format=format, scale=scale, location=obs_location)
        time_obj_converted_format = time.Time(time_obj.to_value("jd"), format="jd", scale=scale, location=obs_location)
        # Perform barycentric correction for scale conversion, will return array of corrected times
        return self._calc_barycentric_time(time_obj_converted_format, obj_location)
    
    def _get_obj_location(self, obj_coords):
        """Creates the Astropy SkyCoord object for BJD time conversions.

        Parameters
        ----------
            obj_coords: (float, float)
                Tuple of the right ascension and declination in degrees of the object being observed.
        
        Returns
        -------
            An Astropy SkyCoord object with the right ascension and declination in degrees.

        Raises
        ------
            ValueError is there is no data for right ascension and/or declination.
        """
        # check if there are objects coords, raise error if not
        if all(elem is None for elem in obj_coords):
            raise ValueError("Recieved None for object right ascension and/or declination. " 
                             "Please enter ICRS coordinate values in degrees for object_ra and object_dec for TransitTime object.")
        # Create SkyCoord object
        return coord.SkyCoord(ra=obj_coords[0], dec=obj_coords[1], unit="deg", frame="icrs")
    
    def _get_obs_location(self, obs_coords, warn):
        """Creates the EarthLocation object for the BJD time conversion.

        Parameters
        ----------
            obs_coords: (float, float)
                Tuple of the longitude and latitude in degrees of the site of observation.
            warn: Boolean
                If True, will raise warnings to the user.
        
        Returns
        -------
            An Astropy EarthLocation object with the latitude and longitude in degrees.
        """
        # Check if there are observatory coords, raise warning and use earth grav center coords if not
        if all(elem is None for elem in obs_coords):
            if warn:
                logging.warning(f"Unable to process observatory coordinates {obs_coords}. "
                                "Using gravitational center of Earth.")
            return coord.EarthLocation.from_geocentric(0., 0., 0., unit=u.m)
        else:
            return coord.EarthLocation.from_geodetic(obs_coords[0], obs_coords[1])
        
    def _sort_data_arrays(self):
        """Sorts all data arrays by ascending epoch.

        Reorders the epochs, mid_times, mid_time_uncertainties, and tra_or_occ arrays based
        on the index order of ascending epochs. This makes sure that all data is order from first 
        observation to most recent observation.
        """
        sorted_indices = np.argsort(self.epochs)
        self.epochs = self.epochs[sorted_indices]
        self.tra_or_occ = self.tra_or_occ[sorted_indices]
        self.mid_times = self.mid_times[sorted_indices]
        self.mid_time_uncertainties = self.mid_time_uncertainties[sorted_indices]
   
    def _validate(self):
        """Checks that all object attributes are of correct types and within value constraints.

        Validates the types of the main data attributes: epochs, mid_times, and mid_time_uncertainties. 
        The following checks are in place:
            * Contained in numpy arrays
            * Epochs are integers
            * Mid times and uncertainties are floats
            * No null or nan values are in the data
            * Epochs, mid times, and uncertainties all contain the same amount of data points
            * All uncertainties are positive
        If the epochs and mid times do not start at zero, the arrays will be shifted to start at zero by 
        subtracting the minimum value of the array from each data point in the array.

        Raises
        ------
            TypeError :
                Error if "epochs", "mid_traisit_times", or "mid_time_uncertainties" are not NumPy arrays.
            ValueError :
                Error if shapes of "epochs", "mid_times", and "mid_time_uncertainties" arrays do not match.
            TypeError :
                Error if values in "epochs" are not ints, values in "mid_times" or "mid_time_uncertainties" are not floats. 
            ValueError :
                Error if "epochs", "mid_times", or "mid_time_uncertainties" contain a NaN (Not-a-Number) value.
            ValueError :
                Error if "mid_time_uncertainties" contains a negative or zero value.
        """
        # Check that all are of type array
        if not isinstance(self.epochs, np.ndarray):
            raise TypeError("The variable `epochs` expected a NumPy array (np.ndarray) but received a different data type")
        if not isinstance(self.mid_times, np.ndarray):
            raise TypeError("The variable `mid_times` expected a NumPy array (np.ndarray) but received a different data type")
        if not isinstance(self.mid_time_uncertainties, np.ndarray):
            raise TypeError("The variable `mid_time_uncertainties` expected a NumPy array (np.ndarray) but received a different data type")
        if not isinstance(self.tra_or_occ, np.ndarray):
            raise TypeError("The variable `tra_or_occ` expected a NumPy array (np.ndarray) but received a different data type")
        # Check that all are same shape
        if self.epochs.shape != self.mid_times.shape != self.mid_time_uncertainties.shape != self.tra_or_occ.shape:
            raise ValueError("Shapes of `epochs`, `mid_times`, `mid_time_uncertainties`, and `tra_or_occ` arrays do not match.")
        # Check that all values in arrays are correct
        if not all(isinstance(value, (int, np.int64, np.int32)) for value in self.epochs):
            raise TypeError("All values in `epochs` must be of type int, numpy.int64, or numpy.int32.")
        if not all(isinstance(value, float) for value in self.mid_times):
            raise TypeError("All values in `mid_times` must be of type float.")
        if not all(isinstance(value, float) for value in self.mid_time_uncertainties):
            raise TypeError("All values in `mid_time_uncertainties` must be of type float.")
        if any(val not in ["tra", "occ"] for val in self.tra_or_occ):
            raise ValueError("The `tra_or_occ` array cannot contain string values other than `tra` or `occ`")
        # Check that there are no null values
        if np.any(np.isnan(self.mid_times)):
            raise ValueError("The `mid_times` array contains NaN (Not-a-Number) values.")
        if np.any(np.isnan(self.mid_time_uncertainties)):
            raise ValueError("The `mid_time_uncertainties` array contains NaN (Not-a-Number) values.")
        # Check that mid_time_uncertainties are positive and non-zero (greater than zero)
        if not np.all(self.mid_time_uncertainties > 0):
            raise ValueError("The `mid_time_uncertainties` array must contain non-negative and non-zero values.")
        # Shift epochs by subtracting the minimum number from everything (new list will start at 0)
        # if self.epochs[0] != 0:
        #     self.epochs = self.epochs.copy() - np.min(self.epochs)
        #     # TODO import warning that we are minimizing their epochs


import sys
sys.path.append(".")
from src.susie.timing_data import TimingData
import unittest
import numpy as np
from numpy.testing import assert_array_equal
import logging
from astropy import time
from astropy import coordinates
from astropy import units as u


# test_epochs = [0, 294, 298, 573, 579, 594, 602, 636, 637, 655, 677, 897, 901, 911, 912, 919, 941, 941, 963, 985, 992, 994, 995, 997, 1015, 1247, 1257, 1258, 1260, 1272, 1287, 1290, 1311, 1312, 1313, 1316, 1317, 1323, 1324, 1333, 1334, 1344, 1345, 1346, 1347, 1357, 1365, 1366, 1585, 1589, 1611, 1619, 1621, 1633, 1637, 1640, 1653, 1661, 1662, 1914, 1915, 1916, 1917, 1937, 1938, 1960, 1964, 1967, 1968, 1969, 1978, 1981, 1991, 1996, 2005, 2012, 2019, 2021, 2022, 2264, 2286, 2288, 2318, 2319, 2331, 2332, 2338, 2339, 2371, 2593, 2634, 2635, 2667, 2668, 2690, 2892, 2910, 2921, 2924, 2942, 2943, 2978, 2979, 2984, 2985, 2988, 2992, 2992, 2997, 2999, 3010, 3017, 3018, 3019, 3217, 3239, 3248, 3260, 3261, 3264, 3306, 3307, 3314, 3316, 3318, 3335, 3335, 3336, 3339, 3341, 3341, 3342, 3342, 3345, 3356, 3570, 3625, 3646, 3657]
# test_mtts = [0.0, 320.8780000000261, 325.24399999994785, 625.3850000002421, 631.933999999892, 648.3059999998659, 657.0360000003129, 694.1440000003204, 695.2370000001974, 714.8820000002161, 738.8940000003204, 979.0049999998882, 983.3710000002757, 994.285000000149, 995.3769999998622, 1003.0160000002943, 1027.0270000002347, 1027.027999999933, 1051.0389999998733, 1075.0509999999776, 1082.691000000108, 1084.8730000001378, 1085.9650000003166, 1088.1480000000447, 1107.7930000000633, 1361.003000000026, 1371.9169999998994, 1373.0079999999143, 1375.191000000108, 1388.2889999998733, 1404.658999999985, 1407.933999999892, 1430.8530000001192, 1431.945000000298, 1433.036000000313, 1436.3100000000559, 1437.4020000002347, 1443.9500000001863, 1445.0419999998994, 1454.8640000000596, 1455.9560000002384, 1466.8700000001118, 1467.9620000002906, 1469.0530000003055, 1470.1450000000186, 1481.058999999892, 1489.7900000000373, 1490.8810000000522, 1729.9020000002347, 1734.2690000003204, 1758.2800000002608, 1767.0109999999404, 1769.194000000134, 1782.2910000002012, 1786.657000000123, 1789.9300000001676, 1804.1189999999478, 1812.851000000257, 1813.942000000272, 2088.9799999999814, 2090.0709999999963, 2091.163000000175, 2092.25400000019, 2114.0819999999367, 2115.1740000001155, 2139.185000000056, 2143.5509999999776, 2146.8250000001863, 2147.916000000201, 2149.0079999999143, 2158.8310000002384, 2162.1049999999814, 2173.0190000003204, 2178.4769999999553, 2188.2990000001155, 2195.939000000246, 2203.5789999999106, 2205.7620000001043, 2206.853000000119, 2470.9769999999553, 2494.9879999998957, 2497.1710000000894, 2529.913000000175, 2531.0049999998882, 2544.1019999999553, 2545.19299999997, 2551.7420000000857, 2552.8330000001006, 2587.7590000000782, 2830.0540000000037, 2874.8020000001416, 2875.8930000001565, 2910.81799999997, 2911.910000000149, 2935.9210000000894, 3156.388000000268, 3176.033999999985, 3188.0389999998733, 3191.313000000082, 3210.9590000002645, 3212.0500000002794, 3250.25, 3251.341000000015, 3256.7990000001155, 3257.8900000001304, 3261.1639999998733, 3265.529000000097, 3265.530999999959, 3270.9870000001974, 3273.1699999999255, 3285.1750000002794, 3292.814999999944, 3293.907000000123, 3294.998000000138, 3511.098999999929, 3535.1099999998696, 3544.933999999892, 3558.0300000002608, 3559.121999999974, 3562.3960000001825, 3608.2349999998696, 3609.3270000000484, 3616.966000000015, 3619.149999999907, 3621.3330000001006, 3639.885000000242, 3639.8870000001043, 3640.978000000119, 3644.253000000026, 3646.435000000056, 3646.435000000056, 3647.526000000071, 3647.526000000071, 3650.8009999999776, 3662.805999999866, 3896.3700000001118, 3956.3980000000447, 3979.31799999997, 3991.323000000324]
# test_mtts_err = [0.00043, 0.00028, 0.00062, 0.00042, 0.00043, 0.00032, 0.00036, 0.00046, 0.00041, 0.00019, 0.00043, 0.00072, 0.00079, 0.00037, 0.00031, 0.0004, 0.0004, 0.00028, 0.00028, 0.00068, 0.00035, 0.00029, 0.00024, 0.00029, 0.00039, 0.00027, 0.00021, 0.00027, 0.00024, 0.00032, 0.00031, 0.00022, 0.00018, 0.00017, 0.00033, 0.00011, 0.0001, 0.00017, 0.00032, 0.00039, 0.00035, 0.00034, 0.00035, 0.00032, 0.00042, 0.00037, 0.00037, 0.00031, 0.00033, 0.00039, 0.0003, 0.0003, 0.0003, 0.0003, 0.00046, 0.00024, 0.00038, 0.00027, 0.00029, 0.00021, 0.0003, 0.00033, 0.00071, 0.00019, 0.00043, 0.00034, 0.00034, 0.00019, 0.00019, 0.00031, 0.00028, 0.00032, 0.0004, 0.00029, 0.00029, 0.00025, 0.00034, 0.00034, 0.00046, 0.00043, 0.00039, 0.00049, 0.00046, 0.00049, 0.00035, 0.00036, 0.00022, 0.0002, 0.00031, 0.00042, 0.00033, 0.00033, 0.00055, 0.00023, 0.00021, 0.00035, 0.00025, 0.00034, 0.00037, 0.00028, 0.00023, 0.00028, 0.00039, 0.00024, 0.00022, 0.00029, 0.00043, 0.00036, 0.00026, 0.00048, 0.00032, 0.0004, 0.00018, 0.00021, 0.00056, 0.00023, 0.0003, 0.00022, 0.00034, 0.00028, 0.00027, 0.00035, 0.00031, 0.00032, 0.00033, 0.0005, 0.00031, 0.00032, 0.00091, 0.00035, 0.00026, 0.00021, 0.00034, 0.00034, 0.00038, 0.0004, 0.00026, 0.0003, 0.00044]
test_epochs = np.array([0, 294, 298, 573])
test_mtts = np.array([2454515.525,2454836.403,2454840.769,2455140.91])
test_mtts_err = np.array([0.00043, 0.00028, 0.00062, 0.00042])
tra_or_occ = np.array(['tra','occ','tra','occ'])
timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ,'utc',object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
expected_mid_times_obj = time.Time(test_mtts,format = 'jd',scale = 'utc')
expected_mid_times_err_obj = time.Time(test_mtts_err,format = 'jd',scale = 'utc')
        
class TestTimingData(unittest.TestCase):
       
    def test_successful_instantiation_jd_tdb_timescale(self):
        """Successful creation of the TimingData object with proper timescale

            Creating the TimingData object with the proper timescale of JD TDB which is
            barycentric Julian Date. Also including uncertainties.
        """
        # Should not get any errors, the epochs and transit times should be the same as they are inputted
        self.timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, time_scale='tdb')
        self.assertIsInstance(self.timing_data, TimingData)  # Check if the object is an instance of TransitTimes
        shifted_epochs = test_epochs - np.min(test_epochs) 
        self.assertTrue(np.array_equal(self.timing_data.epochs, shifted_epochs))  # Check if epochs remain unchanged
        self.assertTrue(np.array_equal(self.timing_data.mid_times, test_mtts))  # Check mid_transit_times
        self.assertTrue(np.array_equal(self.timing_data.mid_time_uncertainties, test_mtts_err))  # Check uncertainties

    def test_s_init_jd_tdb_no_uncertainties(self):
        """ Successful creation of the TimingData object with proper timescale

            Creating the TimingData object with the proper timescale of JD TDB which is
            barycentric Julian Date. Not including uncertainties.
        """
        # Should not get any errors, the epochs and transit times should be the same as they are inputted
        self.timing_data = TimingData('jd', test_epochs, test_mtts, time_scale='tdb')
        self.assertIsInstance(self.timing_data, TimingData )  # Check if the object is an instance of TransitTimes
        shifted_epochs = test_epochs - np.min(test_epochs)
        new_uncertainties = np.full(test_epochs.shape, 0.001)
        self.assertTrue(np.array_equal(self.timing_data.epochs, shifted_epochs))  # Check if epochs remain unchanged
        self.assertTrue(np.array_equal(self.timing_data.mid_times, test_mtts))  # Check mid_transit_times
        self.assertTrue(np.array_equal(self.timing_data.mid_time_uncertainties, new_uncertainties))  # Check uncertainties chage this back!!!

    def test_suc_arrays(self):
        """ Successful test to check that epochs, mid times, mid time uncertainties, and tra_or_occ are all type np.arrays

        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, time_scale='tdb')
        self.assertTrue(isinstance( self.timing_data.epochs, np.ndarray))
        self.assertTrue(isinstance(self.timing_data.mid_times, np.ndarray))
        self.assertTrue(isinstance(self.timing_data.mid_time_uncertainties, np.ndarray))
        self.assertTrue(isinstance(self.timing_data.tra_or_occ, np.ndarray))
   
    def test_us_epochs_arr_type_str(self):
        """ Unsuccessful test to check for numpy array validation for the epochs.
            
            The epochs are strings instead of numpy array and should raise an error.
        """
        string_test_epochs_arr = str(test_epochs)
        with self.assertRaises(TypeError, msg="The variable 'epochs' expected a NumPy array (np.ndarray) but received a different data type"):
             TimingData('jd', string_test_epochs_arr, test_mtts, test_mtts_err, time_scale='tdb')
     
    def test_us_mtts_arr_type_str(self):
        """ Unsuccessful test to check for numpy array validation for the mid times.

            The mid times are strings instead of numpy array and should raise an error.
        """
        string_test_mtts_arr = str(test_mtts)
        with self.assertRaises(TypeError, msg="The variable 'mid_transit_times' expected a NumPy array (np.ndarray) but received a different data type"):
             TimingData('jd', test_epochs, string_test_mtts_arr, test_mtts_err, time_scale='tdb')
    
    def test_us_mtts_err_arr_type_str(self):
        """ Unsuccessful test to check for numpy array validation for the mid time uncertainties.
        
            The mid time uncertainites are strings instead of numpy array and should raise an error.
        """
        string_test_mtts_err_arr = str(test_mtts_err)
        with self.assertRaises(TypeError, msg="The variable 'mid_transit_times_uncertainties' expected a NumPy array (np.ndarray) but received a different data type"):
             TimingData('jd', test_epochs, test_mtts, string_test_mtts_err_arr, time_scale='tdb')  
   
    def test_s_vars_value_types(self):
        """ Successful test to check the correct data type

            Epochs should be integers, mid times should be floats, mid time uncertainties should be floats and tra_or_occ should be strings.
        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, time_scale='tdb')
        self.assertTrue(all(isinstance(value, (int, np.int64)) for value in self.timing_data.epochs))
        self.assertTrue(all(isinstance(value, float) for value in self.timing_data.mid_times))
        self.assertTrue(all(isinstance(value, float) for value in self.timing_data.mid_time_uncertainties))
        self.assertTrue(all(isinstance(value, str) for value in self.timing_data.tra_or_occ))
   
    def test_us_epochs_value_types_float(self):
        """ Unsuccessful test to check for data type validation of the epochs.

            The epochs are floats instead of integers and should raise an error.
        """
        float_test_epochs = test_epochs.astype(float)
        with self.assertRaises(TypeError, msg="All values in 'epochs' must be of type int."):
             TimingData('jd', float_test_epochs, test_mtts, test_mtts_err, time_scale='tdb')
    
    def test_us_mtts_value_types_int(self):
        """ Unsuccessful test to check for data type validation of the mid times.

            The mid times are integers instead of floats and should raise an error.
        """
        int_test_mtts= test_mtts.astype(int)
        with self.assertRaises(TypeError, msg="All values in 'mid_transit_times' must be of type float."):
             TimingData('jd', test_epochs, int_test_mtts, test_mtts_err, time_scale='tdb')
    
    def test_us_mtts_err_value_types_int(self):
        """ Unsuccessful test to check for data type validation of the mid time uncertainties.

            The mid time uncertainties are integers instead of floats and should raise an error.
        """
        int_test_mtts_err= test_mtts_err.astype(int)
        with self.assertRaises(TypeError, msg="All values in 'mid_transit_times_uncertainties' must be of type float."):
             TimingData('jd', test_epochs, test_mtts, int_test_mtts_err, time_scale='tdb')

    # def test_shifted_epochs_zero(self):
    #     """ Successful test to check the shifted epochs function works when the epochs start with 0.

    #         The epochs should remain the same as the array already starts at zero.
    #     """
    #     test_epochs_zero = np.array([0, 294, 298, 573]).astype(int)
    #     shifted_epochs_zero = np.array([0, 294, 298, 573]).astype(int)
    #     self.timing_data =  TimingData('jd', test_epochs_zero, test_mtts, test_mtts_err, time_scale='tdb')
    #     self.assertTrue(np.array_equal(self.timing_data.epochs, shifted_epochs_zero))
    
    # def test_shifted_epochs_pos(self):
    #     """ Successful test to check the shifted epochs function works when the epochs start with a positive number.

    #         The epochs should shift to start with zero.
    #     """
    #     test_epochs_pos = np.array([1, 294, 298, 573]).astype(int)
    #     shifted_epochs_pos = np.array([0, 293, 297, 572]).astype(int)
    #     self.timing_data =  TimingData('jd', test_epochs_pos, test_mtts, test_mtts_err, time_scale='tdb')
    #     self.assertTrue(np.array_equal(self.timing_data.epochs, shifted_epochs_pos))

    # def test_shifted_epochs_neg(self):
    #     """ Successful test to check the shifted epochs function works when the epochs start with a negative number.

    #         The epochs should shift to start with zero.
    #     """
    #     test_epochs_neg = np.array([-1, 294, 298, 573]).astype(int)
    #     shifted_epochs_neg = np.array([0, 295, 299, 574]).astype(int)
    #     self.timing_data =  TimingData('jd', test_epochs_neg, test_mtts, test_mtts_err, time_scale='tdb')
    #     self.assertTrue(np.array_equal(self.timing_data.epochs, shifted_epochs_neg))

   
#<————————————————————————————————————————————————————————————————————————————————————————>
    def test_no_mtts_err(self):
        """ Successful test for when no mid time uncertainties are given.

            If no mid time uncertainties are given the certainties will be an array ones with a data type of floats and the length of the mid times array.
        """
        test_mtts_err = None
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, time_scale='tdb')
        if test_mtts_err is None:
            new_uncertainities= np.ones_like(test_epochs,dtype=float)
            self.assertTrue(np.all(new_uncertainities==np.ones_like(test_epochs,dtype=float)))
        
    def test_mid_transit_err_ones(self):
        """ Successful test for when the mid time uncertainties are an array of ones.
        
        """
        new_test_mtts_err=np.ones_like(test_mtts_err)
        self.timing_data= TimingData('jd', test_epochs, test_mtts, new_test_mtts_err, time_scale='tdb')
        self.assertTrue(np.array_equal(self.timing_data.mid_time_uncertainties,new_test_mtts_err))

    def test_mid_transit_err_neg(self):
        """ Unsuccessful test for when the mid time uncertainties are negative.

            The mid time uncertainties must be postive and will raise an error if the values are negative.
        """
        test_mtts_err_neg= np.array([-0.00043, -0.00028, -0.00062, -0.00042])
        with self.assertRaises(ValueError, msg="The 'mid_transit_times_uncertainties' array must contain non-negative and non-zero values."):
             TimingData('jd', test_epochs, test_mtts, test_mtts_err_neg, time_scale='tdb')  
      
    def test_mid_transit_err_zero(self):
        """ Unsuccessful test for when the mid time uncertainties are zero.

            The mid time uncertianties must be postive and greater than zero and will raise an error if the values are zero.
        """
        test_mtts_err_zero= np.array([0.,0.,0.,0.])
        with self.assertRaises(ValueError, msg="The 'mid_transit_times_uncertainties' array must contain non-negative and non-zero values."):
            TimingData('jd', test_epochs, test_mtts, test_mtts_err_zero, time_scale='tdb')  

    def test_mid_transit_err_self(self):
        """ Successful test for postive and greater than zero mid time uncertainties.

        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, time_scale='tdb')
        self.assertTrue(np.array_equal(self.timing_data.mid_time_uncertainties, test_mtts_err))
   
    def test_variable_shape(self):
        """ Successful test to check that all of the varibles have the same shape.

        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, time_scale='tdb')
        self.assertEqual(test_epochs.shape, test_mtts.shape)
        self.assertEqual(test_epochs.shape, test_mtts_err.shape)
        self.assertEqual(test_epochs.shape, tra_or_occ.shape)

    def test_variable_shape_fail(self):
        """ Unsuccessful test of the varibles shape.

            The epochs, mid times and mid time uncertainties all have different shapes.
        """
        new_test_epochs= np.array([0, 298, 573])  
        new_test_mtts= np.array([0.0, 625.3850000002421])
        new_tra_or_occ = np.array(['tra','tra','occ','occ','occ'])
        with self.assertRaises(ValueError, msg="Shapes of 'epochs', 'mid_transit_times', and 'mid_transit_times_uncertainties' arrays do not match."):
             TimingData('jd', new_test_epochs, new_test_mtts, test_mtts_err, new_tra_or_occ, time_scale='tdb')  
    
    def test_variable_tra_and_occ_shape_fail(self):
        """ Unsuccessful test of the varibles shape.

            The epochs, mid times and mid time uncertainties all have different shapes.
        """
        new_test_epochs= np.array([0, 298, 573])  
        new_test_mtts= np.array([0.0, 625.3850000002421])
        new_tra_or_occ = np.array(['tra','tra','occ','occ','occ'])
        with self.assertRaises(ValueError, msg="Shapes of 'tra_or_occ', 'mid_time_uncertainties', 'mid_times', and 'epochs' arrays do not match."):
             TimingData('jd', new_test_epochs, new_test_mtts, test_mtts_err, new_tra_or_occ, time_scale='tdb')  

    def successful_no_nan_values(self):
        """ Successful test for no Not a Number (NaN) values.

            No NaN values in epochs, mid times and mid times uncertainties.
        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, time_scale='tdb')
        self.assertNotIn(np.nan,test_epochs)
        self.assertNotIn(np.nan,test_mtts)
        self.assertNotIn(np.nan,test_mtts_err)

    def test_mtts_nan(self):
        """ Unsuccessful test to check NaN values in mid times.

            Mid times cannot have any NaN values within the array.
        """
        new_test_mtts=np.array([0., np.nan , 298. ,573.], dtype=float)
        with self.assertRaises(ValueError, msg="The 'mid_transit_times' array contains NaN (Not-a-Number) values."):
             TimingData('jd', test_epochs, new_test_mtts, test_mtts_err, time_scale='tdb')  
    
    def test_mtts_err_nan(self):
        """ Unsuccessful test to check NaN values in mid times uncertainties.

            Mid times uncertainties cannot have any NaN values within the array.
        """
        new_test_mtts_err=np.array([0.00043, np.nan, 0.00062, 0.00042], dtype=float)
        with self.assertRaises(ValueError, msg="The 'mid_transit_times_uncertainties' array contains NaN (Not-a-Number) values."):
             TimingData('jd', test_epochs, test_mtts, new_test_mtts_err, time_scale='tdb')  
  
    # new get obs location test 
    def test_get_obs_location(self):
        """ Successful test to check that the correct observatory coordinates are produced by validate times.

            Testing to see that given a time scale/time format other that jd and tdb the validate times function produces the correct observatory coordinates.
            When given no coordinates the obsevatory coordinates produced should be the gravitational center of the Earth.
        """
        timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, 'utc', object_ra=97.64, object_dec=29.67, observatory_lat=None, observatory_lon=None)
        corrected_times = timing_data._convert_times(test_mtts, 'jd', 'utc',(97.64, 29.67), (None, None))
        obs_location = (0,0)
        expected_obs_location = coordinates.EarthLocation.from_geocentric(0., 0., 0., unit=u.m)
        self.assertTrue(np.allclose(obs_location[0], expected_obs_location.x.value))
        self.assertTrue(np.allclose(obs_location[1], expected_obs_location.y.value))

    def test_obs_geodetic_coords(self):
        """ Successful test to check that the correct observatory coordinates are produced by validate times.

            Testing to see that given a time scale/time format other that jd and tdb the validate times function produces the correct observatory coordinates.
        """
        obs_coords = (-2042896.9,-4149886.2,4376818.8)
        expected_obs_location = coordinates.EarthLocation.from_geodetic(-116.21,43.61)
        self.assertTrue(np.allclose(obs_coords[0], expected_obs_location.x.value))
        self.assertTrue(np.allclose(obs_coords[1], expected_obs_location.y.value))
        self.assertTrue(np.allclose(obs_coords[2], expected_obs_location.z.value))
    # new get obj location test
    def test_get_obj_location(self):
        """ Successful test to check that the correct object coordinates are produced by validate times.

            Testing to see that given a time scale/time format other that jd and tdb the validate times function produces the correct object coordinates.
        """
        timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, 'utc', object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
        obj_coords = (97.64,29.67)
        expected_obj_location = coordinates.SkyCoord(ra=97.64, dec=29.67, unit='deg')
        self.assertEqual(obj_coords[0], expected_obj_location.ra.deg)
        self.assertEqual(obj_coords[1], expected_obj_location.dec.deg)

    def test_validate_times_obj_coords_err(self):
        """ Unsuccessful test to check the object coordinates within the Timing Data.

            The right ascension and declination of the object must be given.
        """
        with self.assertRaises(ValueError, msg="Recieved None for object right ascension and/or declination. Please enter ICRS coordinate values in degrees for object_ra and object_dec for TransitTime object."):
            TimingData('jd', test_epochs, test_mtts, test_mtts_err, time_scale='utc')

    def test_validate_times_obs_coords_logging_err(self):
        """ Test to that that the proper warnings come up for having no observatory coordinates.

            The logging warnings for the time format/time scale and the ICRS coordinates should pop up no matter what.
            The logging warning for the observatory coordinates only pops up when no coordinates are used. The observatory coordinates used becomes the graviational center of the Earth.
        """
        timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, 'utc', object_ra=97.64, object_dec=29.67, observatory_lat=None, observatory_lon=None)
        with self.assertLogs('lumberjack', level='WARNING') as cm:
            logger = logging.getLogger('lumberjack')
            logger.warning('Recieved time format jd and time scale utc. Correcting all times to BJD timing system with TDB time scale. If no time scale is given, default is automatically assigned to UTC. If this is incorrect, please set the time format and time scale for TransitTime object.')
            logger.warning("Unable to process observatory coordinates (None,None). "
                             "Using gravitational center of Earth.")
            logger.warning("Using ICRS coordinates in degrees of RA and Dec (97.64, 29.67) for time correction. "
                        "Using geodetic Earth coordinates in degrees of longitude and latitude (-116.21, 43.6) for time correction.")
        self.assertEqual(cm.output, ['WARNING:lumberjack:Recieved time format jd and time scale utc. Correcting all times to BJD timing system with TDB time scale. If no time scale is given, default is automatically assigned to UTC. If this is incorrect, please set the time format and time scale for TransitTime object.','WARNING:lumberjack:Unable to process observatory coordinates (None,None). Using gravitational center of Earth.','WARNING:lumberjack:Using ICRS coordinates in degrees of RA and Dec (97.64, 29.67) for time correction. Using geodetic Earth coordinates in degrees of longitude and latitude (-116.21, 43.6) for time correction.']) 
        
    def test_validate_times_logging_err(self):
        """ Test to that that the proper warnings come up for having observatory coordinates.

            The logging warnings for the time format/time scale and the ICRS coordinates should pop up no matter what.
        """
        timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ, 'utc', object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
        with self.assertLogs('lumberjack', level='WARNING') as cm:
            logger = logging.getLogger('lumberjack')
            logger.warning('Recieved time format jd and time scale utc. Correcting all times to BJD timing system with TDB time scale. If no time scale is given, default is automatically assigned to UTC. If this is incorrect, please set the time format and time scale for TransitTime object.')
            logger.warning('Using ICRS coordinates in degrees of RA and Dec (97.64, 29.67) for time correction. Using geodetic Earth coordinates in degrees of longitude and latitude (-116.21, 43.6) for time correction.')
        self.assertEqual(cm.output, ['WARNING:lumberjack:Recieved time format jd and time scale utc. Correcting all times to BJD timing system with TDB time scale. If no time scale is given, default is automatically assigned to UTC. If this is incorrect, please set the time format and time scale for TransitTime object.','WARNING:lumberjack:Using ICRS coordinates in degrees of RA and Dec (97.64, 29.67) for time correction. Using geodetic Earth coordinates in degrees of longitude and latitude (-116.21, 43.6) for time correction.']) 
    
    

    def test_calc_bary_time_instantiation(self):
        """ Successful test for instantiation of timing data.

        """
        self.timing_data =  TimingData('jd', test_epochs, test_mtts, test_mtts_err, object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
        self.assertIsInstance(self.timing_data,  TimingData)
    
    def test_calc_bary_time_uncertainties(self):
        """ Successful test to check correct process when the uncertainties are ones for the barycentric time function.

            If the uncertainties are ones, there is no correction done as they are just placeholder values.
        """
        self.timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
        obs_location = coordinates.EarthLocation(lat = 43.60, lon = -116.21)
        time_obj = time.Time(test_mtts_err, format='jd', scale='utc', location=obs_location)
        obj_location = coordinates.SkyCoord(ra = 97.6,dec = 29.67, unit = 'deg')
        obs_location = coordinates.EarthLocation(lat = 43.60, lon = 116.21)
        expected_result = np.array([0.00354448, 0.00339449, 0.00373446, 0.00353448])
        actual_result = self.timing_data._calc_barycentric_time(time_obj,obj_location)
        self.assertTrue(np.allclose(expected_result, actual_result,rtol=1e-05, atol=1e-08))

    def test_bary_time_corrected_time_vals(self):
        """ Successful test to check that the corrected time values are the correct values.

            Within the barycentric time function the corrected time time values are the values that have been fully corrected for the barycentric light correction.
        """
        test_epochs = np.array([0, 294, 298, 573])
        test_mtts = np.array([0.0, 320.8780000000261, 325.24399999994785, 625.3850000002421])
        timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, object_ra=97.64, object_dec=29.67, observatory_lat=43.60, observatory_lon=-116.21)
        obs_location = coordinates.EarthLocation(lat = 43.60, lon = -116.21)
        time_obj = time.Time(test_mtts,format = 'jd',scale = 'utc',location = obs_location)
        obj_location = coordinates.SkyCoord(ra = 97.6,dec = 29.67, unit = 'deg')
        expected_result = np.array([3.11451337e-03,3.20883762e+02,3.25249628e+02,6.25389263e+02])
        actual_result = timing_data._calc_barycentric_time(time_obj, obj_location)
        self.assertTrue((np.allclose(expected_result,actual_result, rtol=1e-05, atol=1e-08)))

    def test_ltt_bary(self):
        """ Successful check for the correct barycentric time correction.

            Within the barycentric time function the ltt bary values are the amount needed to adjust the values by for the barycenter.
        """
        time_obj = time.Time(test_mtts, format='jd', scale='utc')
        obj_location = coordinates.SkyCoord(ra=97.6, dec=29.67, unit='deg')
        obs_location = coordinates.EarthLocation(lat=43.60, lon=-116.21)
        time_obj.location = obs_location
        actual_ltt_bary = time_obj.light_travel_time(obj_location).value
        expected_ltt_bary = np.array([0.00344345,0.00561672,0.00554016,0.00339291])
        self.assertTrue((np.allclose(expected_ltt_bary, actual_ltt_bary, rtol=1e-05, atol=1e-08)))

    def test_convert_timing_unc(self):
        """ Sucessful tests for the mid time uncertainty conversions.

            The mid time uncertainties are calulated using the upper limit (mid times + uncertainties) and lower limits (mid times - uncertainties) only if the scale/format needs to be converted.
        """
        self.timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ = None, time_scale='tdb')
        expected_result = np.array([0.00060806, 0.00039597, 0.00087679, 0.00059402])
        result = self.timing_data._convert_timing_uncertainties(test_mtts, test_mtts_err, 'jd', 'tdb', (97.64,29.67), (-116.21,43.60))
        self.assertTrue((np.allclose(expected_result,result, rtol=1e-05, atol=1e-08)))


    def test_tra_or_occ_None(self):
        """ Successful test for when the tra_or_occ equals None.

            If there is no tra_or_occ data given an array of 'tra' are produced with the length of the epoch array.
        """
        self.timing_data = TimingData('jd', test_epochs, test_mtts, test_mtts_err, tra_or_occ = None, time_scale='tdb')
        expected_result = np.array(['tra','tra','tra','tra'])
        result = self.timing_data.tra_or_occ
        assert_array_equal(expected_result, result)
    
    def test_only_tra_or_occ_value(self):
        """ Unsuccessful test to check if the tra_or_occ array contains values other than 'tra' or 'occ'.

        """
        not_tra_or_occ = np.array(['tra','occ','trac','oc'])
        with self.assertRaises(ValueError, msg="tra_or_occ array cannot contain string values other than 'tra' or 'occ'"):
             TimingData('jd', test_epochs, test_mtts, test_mtts_err,not_tra_or_occ, time_scale='tdb')  

    def test_tra_or_occ_array(self):
        """ Unsuccessful test to check if the tra_or_occ varible is a numpy array.

        """
        not_tra_or_occ = str(tra_or_occ)
        with self.assertRaises(TypeError, msg= "The variable 'tra_or_occ' expected a NumPy array (np.ndarray) but received a different data type"):
             TimingData('jd', test_epochs, test_mtts, test_mtts_err,not_tra_or_occ, time_scale='tdb')  

    def test_tra_or_occ_shape(self):
        """ Unsuccessful test to check the length of the tra_or_occ array.

            The tra_or_occ array must be the same length as the epochs. mid times and mid time uncertainties arrays.
        """
        not_test_epochs = np.array([0,1])
        not_test_mtts = np.array([0.,1.0,3.0,4.0,5.0])
        not_tra_or_occ = np.array(['occ','tra','occ'])
        with self.assertRaises(ValueError, msg= "Shapes of 'tra_or_occ', 'mid_time_uncertainties', and 'mid_times' arrays do not match."):
             TimingData('jd', not_test_epochs, not_test_mtts, test_mtts_err,not_tra_or_occ, time_scale='tdb')  

    def test_tra_or_occ_no_null(self):
        """ Unsuccessful test to check if any null values are in the tra_or_occ array.

            The data values cannot contain any null values of an error will be rasied if not.
        """
        not_tra_or_occ = np.array(['tra','occ', None, None])
        with self.assertRaises(ValueError, msg = "The 'tra_or_occ' array contains NaN (Not-a-Number) values."):
             TimingData('jd', test_epochs, test_mtts, test_mtts_err,not_tra_or_occ, time_scale='tdb')  
    

    # sort arrays
    def test_sort_arrays(self):
        """ Successful test to sort arrays.

            Arrays will be sorted by the first epoch. This occurs after the minimization of the epochs so zero should be first.
        """
        test_epochs_reordered = np.array([294, 0, 298, 573])
        test_mtts_reordered = np.array([2454836.403,2454515.525,2454840.769,2455140.91])
        test_mtts_err_reordered = np.array([ 0.00028,0.00043, 0.00062, 0.00042])
        tra_or_occ_reordered = np.array(['occ','tra','tra','occ'])
        timing_data_reordered = TimingData('jd', test_epochs_reordered, test_mtts_reordered, test_mtts_err_reordered, tra_or_occ_reordered, time_scale='tdb')
        expected_epochs = np.array([0, 294, 298, 573])
        expected_mtts = np.array([2454515.525,2454836.403,2454840.769,2455140.91])
        expected_mtts_err = np.array([0.00043, 0.00028, 0.00062, 0.00042])
        expected_tra_or_occ = np.array(['tra','occ','tra','occ'])
        self.assertTrue((np.allclose(timing_data_reordered.epochs, expected_epochs, rtol=1e-05, atol=1e-08)))
        self.assertTrue((np.allclose(timing_data_reordered.mid_times,expected_mtts, rtol=1e-05, atol=1e-08)))
        self.assertTrue((np.allclose(timing_data_reordered.mid_time_uncertainties, expected_mtts_err, rtol=1e-05, atol=1e-08)))
        assert_array_equal(timing_data_reordered.tra_or_occ, expected_tra_or_occ)


    
    # def test_successful_instantiation_jd_no_timescale(self):
    #     transit_times = TransitTimes('jd', )
    # def test_successful_instantiation_jd_non_tdb_timescale(self):
    #     transit_times = TransitTimes('jd', )
    # def test_successful_instantiation_non_jd_tdb_timescale(self):
    #     transit_times = TransitTimes('mjd', time_scale='tdb')
    # def test_successful_instantiation_non_jd_no_timescale(self):
    #     transit_times = TransitTimes('', )
    # def test_successful_instantiation_non_jd_non_tdb_timescale(self):
    #     transit_times = TransitTimes('', )
    # # Test instantiating with ra/dec and without ra/dec vals (and only one val)
    # # Test instantiating
    # def test_no_format(self):
    #     transit_times = TransitTimes()
    # def test_no_obj_coords(self):
    #     transit_times = TransitTimes()
    # def test_all_data_success():
    #     pass

    # def test_all_data_fail():
    #     pass

    # def test_no_uncertainties():
    #     pass
if __name__ == "__main__":
    unittest.main()


import pytz
import numpy as np
import pandas as pd
from astropy.units import Quantity
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord
from astroplan import FixedTarget, Observer, EclipsingSystem, AtNightConstraint, AltitudeConstraint, is_event_observable
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive

class Campaign():
    def __init__(self):
        pass

    def _query_nasa_exoplanet_archive(self, obj_name, ra=None, dec=None, select_query=None):
        """Queries the NASA Exoplanet Archive for system parameters.

        Parameters
        ----------
            obj_name: str
                The name of the exoplanet object.
            ra: float (Optional)
                The right ascension of the object to observe in the sky (most likely a planet or star).
            dec: float (Optional)
                The declination of the object to observe in the sky (most likely a planet or star).
            select_query: str
                The select query string. For examples please see the Astroquery documentation.
            
        Returns
        -------
            obj_data: An astropy Table object
                The table of data returned from the NASA Exoplanet Archive query. For more information on 
                how to work with these tables, please see the Astroquery NASA Exoplanet Archive documentation.
        """
        # Get object data
        obj_data = None
        if obj_name is not None:
            if select_query:
                obj_data = NasaExoplanetArchive.query_object(obj_name, select=select_query)
            else:
                obj_data = NasaExoplanetArchive.query_object(obj_name)
        elif ra is not None and dec is not None:
            if select_query:
                obj_data = NasaExoplanetArchive.query_region(
                    table="pscomppars", coordinates=SkyCoord(ra=ra*u.deg, dec=dec*u.deg),
                    radius=1.0*u.deg, select=select_query)
            else:
                obj_data = NasaExoplanetArchive.query_region(
                    table="pscomppars", coordinates=SkyCoord(ra=ra*u.deg, dec=dec*u.deg), radius=1.0*u.deg)
        else:
            raise ValueError("Object must be specified with either (1) a recognized object name in the NASA Exoplanet Archive or (2) right ascension and declination in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec.")
        # Check that the returned data is not empty
        if obj_data is not None and len(obj_data) > 0:
            return obj_data
        else:
            if obj_name is not None:
                raise ValueError(f"Nothing found for {obj_name} in the NASA Exoplanet Archive. Please check that your object is accepted and contains data on the archive's homepage.")
            elif ra is not None and dec is not None:
                raise ValueError(f"Nothing found for the coordinates {ra}, {dec} in the NASA Exoplanet Archive. Please check that your values are correct and are in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec.")
    
    def _get_eclipse_duration(self, obj_name, ra=None, dec=None):
        """Queries the NASA Exoplanet Archive for system parameters used in eclipse duration calculation.

        Parameters
        ----------
            obj_name: str
                The name of the exoplanet object.
            ra: float (Optional)
                The right ascension of the object to observe in the sky (most likely a planet or star).
            dec: float (Optional)
                The declination of the object to observe in the sky (most likely a planet or star).
            
        Returns
        -------
        """
        nea_data = self._query_nasa_exoplanet_archive(obj_name, ra, dec, select_query="pl_trandur")
        for val in nea_data["pl_trandur"]:
            if not(np.isnan(val)):
                val_to_store = val
                if isinstance(val, Quantity) and hasattr(val, 'mask'):
                    # If the value is masked, just store value
                    val_to_store = val.value
                return val_to_store * u.hour
            
    def _validate_obs_start_time(self, time_str):
        value_err_msg = "obs_start_time must be in the format YYYY-MM-DD. For example, 2024-10-29."
        if len(time_str) != 10:
            return ValueError(value_err_msg)
        if not (time_str[0:4].isdigit() and time_str[5:7].isdigit() and time_str[8:10].isdigit()):
            return ValueError(value_err_msg)
        if not (time_str[4] == '-' and time_str[7] == '-'):
            return ValueError(value_err_msg)
        
    def _validate_observing_schedule_params(self, observer, target, obs_start_time):
        self._validate_obs_start_time(obs_start_time)
        if not isinstance(observer, Observer):
            return TypeError("observer parameter must be an Astroplan Observer object. See the Astroplan documentation for more information: https://astroplan.readthedocs.io/en/latest/api/astroplan.Observer.html")
        if not isinstance(target, FixedTarget):
            return TypeError("target parameter must be an Astroplan FixedTarget object. See the Astroplan documentation for more information: https://astroplan.readthedocs.io/en/latest/api/astroplan.Target.html")
        
    def _create_observing_schedule_dataframe(self, transits, occultations, time_format):
        transit_df = pd.DataFrame(transits)
        occultation_df = pd.DataFrame(occultations)
        if time_format == "datetime":
            transit_df = transit_df.map(lambda dt: dt.strftime("%Y-%m-%d %H:%M:%S %z"))
            occultation_df = occultation_df.map(lambda dt: dt.strftime("%Y-%m-%d %H:%M:%S %z"))
        transit_df = transit_df.rename(columns={0: "ingress", 1: "egress"})
        occultation_df = occultation_df.rename(columns={0: "ingress", 1: "egress"})
        transit_df["type"] = "transit"
        occultation_df["type"] = "occultation"
        final_df = pd.concat([transit_df, occultation_df], ignore_index=True)
        sorted_df = final_df.sort_values(by="ingress", ascending=True)
        return sorted_df

    def create_observer_obj(self, timezone, name, longitude=None, latitude=None, elevation=0.0):
        """Creates the Astroplan Observer object.

        Parameters
        ----------
            timezone: str
                The local timezone. If a string, it will be passed through pytz.timezone() to produce the timezone object.
            name: str
                The name of the observer's location. This can either be a registered Astropy site
                name (get the latest site names with `EarthLocation.get_site_names()`), which will
                return the latitude, longitude, and elevation of the site OR it can be a custom name
                to keep track of your Observer object.
            latitude: float (Optional)
                The latitude of the observer's location on Earth.
            longitude: float (Optional)
                The longitude of the observer's location on Earth.
            elevation: float (Optional)
                The elevation of the observer's location on Earth.

        Returns
        -------
            The Astroplan Observer object.
        
        Raises
        ------
            ValueError if neither coords nor name are given.
        """
        observer = None
        if longitude is not None and latitude is not None:
            # There are valid vals for lon and lat
            observer = Observer(longitude=longitude*u.deg, latitude=latitude*u.deg, elevation=elevation*u.m, timezone=timezone)
            if name is not None:
                observer.name = name
        elif name is not None:
            # No vals for lon and lat, use site name
            observer = Observer.at_site(name, timezone=timezone)
        else:
            # No coords or site name given, raise error
            raise ValueError("Observatory location must be specified with either (1) a site name specified by astropy.coordinates.EarthLocation.get_site_names() or (2) latitude and longitude in degrees as accepted by astropy.coordinates.Latitude and astropy.coordinates.Latitude.")
        return observer
            
    def create_target_obj(self, name, ra=None, dec=None):
        """Creates the Astroplan FixedTarget object.

        Parameters
        ----------
            coords: tuple(float, float) (Optional)
                The right ascension and declination of the object in the sky (most likely the planet's host star).
            name: str
                The name of the exoplanet host star. This can either be a registered object name, which will query
                a CDS name resolver (see the `Astroplan Target Documentation <https://astroplan.readthedocs.io/en/latest/api/astroplan.Target.html>`_ 
                for more information on this) OR it can be a custom name to keep track of your FixedTarget object.
            ra: float (Optional)
                The right ascension of the object to observe in the sky (most likely a planet or star).
            dec: float (Optional)
                The declination of the object to observe in the sky (most likely a planet or star).
        
        Returns
        -------
            The Astroplan FixedTarget object.

        Raises
        ------
            ValueError if neither coords nor name are given.
        """
        target = None
        if ra is not None and dec is not None:
            # There are valid vals for ra and dec
            skycoord = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)
            target = FixedTarget(coord=skycoord)
            if name is not None:
                target.name = name
        elif name is not None:
            # No vals for ra & dec, query by object name
            target = FixedTarget.from_name(name)
        else:
            # Neither ra & dec or name given, raise error
            raise ValueError("Object location must be specified with either (1) an valid object name or (2) right ascension and declination in degrees as accepted by astropy.coordinates.ra and astropy.coordinates.dec.")
        return target
    
    def get_observing_schedule(self, primary_eclipse_time, orbital_period, timezone, observer, target, n_transits, n_occultations, obs_start_time, exoplanet_name=None, eclipse_duration=None, csv_filename=None, return_time_format="datetime"):
        """Returns a list of observable future transits for the target object

        Parameters
        ----------
            primary_eclipse_time: Astropy Time obj
                The most recent mid-transit time with Julian Date time format.
            orbital_period: float
                The orbital period of the exoplanet in days.
            timezone: str
                The local timezone. If a string, it will be passed through `pytz.timezone()` to produce the timezone object.
            observer: Astroplan Observer obj
                An Astroplan Observer object holding information on the observer's Earth location. Can be created 
                through the `create_observer_obj` method, or can be manually created. See the `Astroplan Observer Documentation <https://astroplan.readthedocs.io/en/latest/api/astroplan.Observer.html>`_
                for more information.
            target: Astroplan FixedTarget obj
                An Astroplan FixedTarget object holding information on the object observed. Can be created through the 
                `create_target_obj` method, or can be manually created. See the `Astroplan Target Documentation <https://astroplan.readthedocs.io/en/latest/api/astroplan.Target.html>`_
                for more information.
            n_transits: int
                The number of transits to initially request. This will be filtered down by what is observable from
                the Earth location.
            n_occultations: int
                The number of occultations to initially request. This will be filtered down by what is observable from
                the Earth location.
            obs_start_time: str
                Time at which you would like to start looking for eclipse events. In the format YYYY-MM-DD. For
                example, if you would like to find eclipses happening after October 1st, 2024, the format would
                be "2024-10-01".
            exoplanet_name: str (Optional)
                The name of the exoplanet. Used to query the NASA Exoplanet Archive for transit duration. If 
                no name is provided, the right ascension and declination from the FixedTarget object will be used. 
                Can also provide the transit duration manually instead using the `eclipse_duration` parameter.
            eclipse_duration: float (Optional)
                The full duration of the exoplanet transit from ingress to egress in hours. If not given, will calculate
                using either provided system parameters or parameters pulled from the NASA Exoplanet Archive.
            csv_filename: str (Optional)
                If given, will save the returned schedule dataframe as a CSV file.
            csv_time_format: str (Optional)
                The time format given in the returned CSV file. Default of datetime format "%Y-%m-%d %H:%M:%S %z".
        """
        # Validate some things before continuing
        self._validate_observing_schedule_params(observer, target, obs_start_time)
        # Grab the most recent mid transit time
        primary_eclipse_time = Time(primary_eclipse_time, format='jd')
        # Pull orbital period from the model data
        orbital_period = orbital_period * u.day
        if eclipse_duration == None:
            # If not given, query the archive for it
            eclipse_duration = self._get_eclipse_duration(exoplanet_name, target.ra, target.dec)
        # Create EclipsingSystem object
        eclipsing_system = EclipsingSystem(primary_eclipse_time=primary_eclipse_time,
                                orbital_period=orbital_period, duration=eclipse_duration)
        # Set the observational parameters
        # Time to start looking
        obs_time = Time(f"{obs_start_time} 00:00")
        # Grab the number of transits and occultations asked for
        ing_egr_transits = eclipsing_system.next_primary_ingress_egress_time(obs_time, n_eclipses=n_transits)
        ing_egr_occultations = eclipsing_system.next_secondary_ingress_egress_time(obs_time, n_eclipses=n_occultations)
        # We need to check if the events are observable
        constraints = [AtNightConstraint.twilight_civil(), AltitudeConstraint(min=30*u.deg)]
        transits_bool = is_event_observable(constraints, observer, target, times_ingress_egress=ing_egr_transits)
        occultations_bool = is_event_observable(constraints, observer, target, times_ingress_egress=ing_egr_occultations)
        observable_transits = ing_egr_transits[transits_bool[0]]
        observable_occultations = ing_egr_occultations[occultations_bool[0]]
        # Change to their given timezone
        if return_time_format == "datetime":
            tz = pytz.timezone(timezone)
            observable_transits = observable_transits.to_datetime(timezone=tz)
            observable_occultations = observable_occultations.to_datetime(timezone=tz)
        # Create dataframe from this
        schedule_df = self._create_observing_schedule_dataframe(observable_transits, observable_occultations, return_time_format)
        if csv_filename is not None:
            schedule_df.to_csv(csv_filename, index=False)
        return schedule_df